# pylint: disable=too-many-lines,line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import copy
import functools
import importlib
import json
import logging
import os
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
from urllib.parse import urlparse

from azure.ai.projects import _types
from azure.ai.projects.models import AgentRunStream, AsyncAgentRunStream, _models
from azure.ai.projects.models._enums import AgentsApiResponseFormatMode, MessageRole, RunStepStatus
from azure.ai.projects.models import (
    MessageAttachment,
    MessageDeltaChunk,
    MessageIncompleteDetails,
    RunStep,
    RunStepDeltaChunk,
    RunStepError,
    RunStepFunctionToolCall,
    RunStepToolCallDetails,
    RunStepCodeInterpreterToolCall,
    RunStepBingGroundingToolCall,
    ThreadMessage,
    ThreadRun,
    ToolDefinition,
    ToolOutput,
    ToolResources,
)
from azure.ai.projects.models._patch import AgentEventHandler, AsyncAgentEventHandler, ToolSet
from azure.ai.projects.telemetry.agents._utils import (
    AZ_AI_AGENT_SYSTEM,
    ERROR_TYPE,
    GEN_AI_AGENT_DESCRIPTION,
    GEN_AI_AGENT_ID,
    GEN_AI_AGENT_NAME,
    GEN_AI_EVENT_CONTENT,
    GEN_AI_MESSAGE_ID,
    GEN_AI_MESSAGE_STATUS,
    GEN_AI_RESPONSE_MODEL,
    GEN_AI_SYSTEM,
    GEN_AI_SYSTEM_MESSAGE,
    GEN_AI_THREAD_ID,
    GEN_AI_THREAD_RUN_ID,
    GEN_AI_THREAD_RUN_STATUS,
    GEN_AI_USAGE_INPUT_TOKENS,
    GEN_AI_USAGE_OUTPUT_TOKENS,
    GEN_AI_RUN_STEP_START_TIMESTAMP,
    GEN_AI_RUN_STEP_END_TIMESTAMP,
    GEN_AI_RUN_STEP_STATUS,
    ERROR_MESSAGE,
    OperationName,
    start_span,
)
from azure.core import CaseInsensitiveEnumMeta  # type: ignore
from azure.core.settings import settings
from azure.core.tracing import AbstractSpan

_Unset: Any = object()

try:
    # pylint: disable = no-name-in-module
    from opentelemetry.trace import Span, StatusCode

    _tracing_library_available = True
except ModuleNotFoundError:
    _tracing_library_available = False


__all__ = [
    "AIAgentsInstrumentor",
]


_agents_traces_enabled: bool = False
_trace_agents_content: bool = False


class TraceType(str, Enum, metaclass=CaseInsensitiveEnumMeta):  # pylint: disable=C4747
    """An enumeration class to represent different types of traces."""

    AGENTS = "Agents"


class AIAgentsInstrumentor:
    """
    A class for managing the trace instrumentation of AI Agents.

    This class allows enabling or disabling tracing for AI Agents.
    and provides functionality to check whether instrumentation is active.

    """

    def __init__(self):
        if not _tracing_library_available:
            raise ModuleNotFoundError(
                "Azure Core Tracing Opentelemetry is not installed. "
                "Please install it using 'pip install azure-core-tracing-opentelemetry'"
            )
        # In the future we could support different versions from the same library
        # and have a parameter that specifies the version to use.
        self._impl = _AIAgentsInstrumentorPreview()

    def instrument(self, enable_content_recording: Optional[bool] = None) -> None:
        """
        Enable trace instrumentation for AI Agents.

        :param enable_content_recording: Whether content recording is enabled as part
          of the traces or not. Content in this context refers to chat message content
          and function call tool related function names, function parameter names and
          values. True will enable content recording, False will disable it. If no value
          is provided, then the value read from environment variable
          AZURE_TRACING_GEN_AI_CONTENT_RECORDING_ENABLED is used. If the environment variable
          is not found, then the value will default to False. Please note that successive calls
          to instrument will always apply the content recording value provided with the most
          recent call to instrument (including applying the environment variable if no value is
          provided and defaulting to false if the environment variable is not found), even if
          instrument was already previously called without uninstrument being called in between
          the instrument calls.
        :type enable_content_recording: bool, optional

        """
        self._impl.instrument(enable_content_recording)

    def uninstrument(self) -> None:
        """
        Remove trace instrumentation for AI Agents.

        This method removes any active instrumentation, stopping the tracing
        of AI Agents.
        """
        self._impl.uninstrument()

    def is_instrumented(self) -> bool:
        """
        Check if trace instrumentation for AI Agents is currently enabled.

        :return: True if instrumentation is active, False otherwise.
        :rtype: bool
        """
        return self._impl.is_instrumented()

    def is_content_recording_enabled(self) -> bool:
        """This function gets the content recording value.

        :return: A bool value indicating whether content recording is enabled.
        :rtype: bool
        """
        return self._impl.is_content_recording_enabled()


class _AIAgentsInstrumentorPreview:
    # pylint: disable=R0904
    """
    A class for managing the trace instrumentation of AI Agents.

    This class allows enabling or disabling tracing for AI Agents.
    and provides functionality to check whether instrumentation is active.
    """

    def _str_to_bool(self, s):
        if s is None:
            return False
        return str(s).lower() == "true"

    def instrument(self, enable_content_recording: Optional[bool] = None):
        """
        Enable trace instrumentation for AI Agents.

        :param enable_content_recording: Whether content recording is enabled as part
        of the traces or not. Content in this context refers to chat message content
        and function call tool related function names, function parameter names and
        values. True will enable content recording, False will disable it. If no value
        is provided, then the value read from environment variable
        AZURE_TRACING_GEN_AI_CONTENT_RECORDING_ENABLED is used. If the environment variable
        is not found, then the value will default to False.

        :type enable_content_recording: bool, optional
        """
        if enable_content_recording is None:
            var_value = os.environ.get("AZURE_TRACING_GEN_AI_CONTENT_RECORDING_ENABLED")
            enable_content_recording = self._str_to_bool(var_value)
        if not self.is_instrumented():
            self._instrument_agents(enable_content_recording)
        else:
            self._set_enable_content_recording(enable_content_recording=enable_content_recording)

    def uninstrument(self):
        """
        Disable trace instrumentation for AI Agents.

        This method removes any active instrumentation, stopping the tracing
        of AI Agents.
        """
        if self.is_instrumented():
            self._uninstrument_agents()

    def is_instrumented(self):
        """
        Check if trace instrumentation for AI Agents is currently enabled.

        :return: True if instrumentation is active, False otherwise.
        :rtype: bool
        """
        return self._is_instrumented()

    def set_enable_content_recording(self, enable_content_recording: bool = False) -> None:
        """This function sets the content recording value.

        :param enable_content_recording: Indicates whether tracing of message content should be enabled.
                                    This also controls whether function call tool function names,
                                    parameter names and parameter values are traced.
        :type enable_content_recording: bool
        """
        self._set_enable_content_recording(enable_content_recording=enable_content_recording)

    def is_content_recording_enabled(self) -> bool:
        """This function gets the content recording value.

        :return: A bool value indicating whether content tracing is enabled.
        :rtype bool
        """
        return self._is_content_recording_enabled()

    def _set_attributes(self, span: "AbstractSpan", *attrs: Tuple[str, Any]) -> None:
        for attr in attrs:
            key, value = attr
            if value is not None:
                span.add_attribute(key, value)

    def _parse_url(self, url):
        parsed = urlparse(url)
        server_address = parsed.hostname
        port = parsed.port
        return server_address, port

    def _remove_function_call_names_and_arguments(self, tool_calls: list) -> list:
        tool_calls_copy = copy.deepcopy(tool_calls)
        for tool_call in tool_calls_copy:
            if "function" in tool_call:
                if "name" in tool_call["function"]:
                    del tool_call["function"]["name"]
                if "arguments" in tool_call["function"]:
                    del tool_call["function"]["arguments"]
                if not tool_call["function"]:
                    del tool_call["function"]
        return tool_calls_copy

    def _create_event_attributes(
        self,
        thread_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        thread_run_id: Optional[str] = None,
        message_id: Optional[str] = None,
        message_status: Optional[str] = None,
        run_step_status: Optional[str] = None,
        created_at: Optional[datetime] = None,
        completed_at: Optional[datetime] = None,
        cancelled_at: Optional[datetime] = None,
        failed_at: Optional[datetime] = None,
        run_step_last_error: Optional[RunStepError] = None,
        usage: Optional[_models.RunStepCompletionUsage] = None,
    ) -> Dict[str, Any]:
        attrs: Dict[str, Any] = {GEN_AI_SYSTEM: AZ_AI_AGENT_SYSTEM}
        if thread_id:
            attrs[GEN_AI_THREAD_ID] = thread_id

        if agent_id:
            attrs[GEN_AI_AGENT_ID] = agent_id

        if thread_run_id:
            attrs[GEN_AI_THREAD_RUN_ID] = thread_run_id

        if message_id:
            attrs[GEN_AI_MESSAGE_ID] = message_id

        if message_status:
            attrs[GEN_AI_MESSAGE_STATUS] = self._status_to_string(message_status)

        if run_step_status:
            attrs[GEN_AI_RUN_STEP_STATUS] = self._status_to_string(run_step_status)

        if created_at:
            if isinstance(created_at, datetime):
                attrs[GEN_AI_RUN_STEP_START_TIMESTAMP] = created_at.isoformat()
            else:
                # fallback in case integer or string gets passed
                attrs[GEN_AI_RUN_STEP_START_TIMESTAMP] = str(created_at)

        end_timestamp = None
        if completed_at:
            end_timestamp = completed_at
        elif cancelled_at:
            end_timestamp = cancelled_at
        elif failed_at:
            end_timestamp = failed_at

        if isinstance(end_timestamp, datetime):
            attrs[GEN_AI_RUN_STEP_END_TIMESTAMP] = end_timestamp.isoformat()
        elif end_timestamp:
            # fallback in case int or string gets passed
            attrs[GEN_AI_RUN_STEP_END_TIMESTAMP] = str(end_timestamp)

        if run_step_last_error:
            attrs[ERROR_MESSAGE] = run_step_last_error.message
            attrs[ERROR_TYPE] = run_step_last_error.code

        if usage:
            attrs[GEN_AI_USAGE_INPUT_TOKENS] = usage.prompt_tokens
            attrs[GEN_AI_USAGE_OUTPUT_TOKENS] = usage.completion_tokens

        return attrs

    def add_thread_message_event(
        self, span, message: ThreadMessage, usage: Optional[_models.RunStepCompletionUsage] = None
    ) -> None:
        content_body = {}
        if _trace_agents_content:
            for content in message.content:
                typed_content = content.get(content.type, None)
                if typed_content:
                    content_details = {"value": self._get_field(typed_content, "value")}
                    annotations = self._get_field(typed_content, "annotations")
                    if annotations:
                        content_details["annotations"] = [a.as_dict() for a in annotations]
                    content_body[content.type] = content_details

        self._add_message_event(
            span,
            self._get_role(message.role),
            content_body,
            attachments=message.attachments,
            thread_id=message.thread_id,
            agent_id=message.agent_id,
            message_id=message.id,
            thread_run_id=message.run_id,
            message_status=message.status,
            incomplete_details=message.incomplete_details,
            usage=usage,
        )

    def _process_tool_calls(self, step: RunStep) -> List[Dict[str, Any]]:
        """
        Helper method to process tool calls and return a list of tool call dictionaries.

        :param step: The run step containing tool call details to be processed.
        :type step: RunStep
        :return: A list of dictionaries, each representing a processed tool call.
        :rtype: List[Dict[str, Any]]
        """
        tool_calls = []
        tool_call: Dict[str, Any] = {}
        for t in cast(RunStepToolCallDetails, step.step_details).tool_calls:
            if not _trace_agents_content:
                tool_call = {
                    "id": t.id,
                    "type": t.type,
                }
            elif isinstance(t, RunStepFunctionToolCall):
                try:
                    parsed_arguments = json.loads(t.function.arguments)
                except json.JSONDecodeError:
                    parsed_arguments = {}

                tool_call = {
                    "id": t.id,
                    "type": t.type,
                    "function": {
                        "name": t.function.name,
                        "arguments": parsed_arguments,
                    },
                }
            elif isinstance(t, RunStepCodeInterpreterToolCall):
                tool_call = {
                    "id": t.id,
                    "type": t.type,
                    "code_interpreter": {
                        "input": t.code_interpreter.input,
                        "outputs": [output.as_dict() for output in t.code_interpreter.outputs],
                    },
                }
            elif isinstance(t, RunStepBingGroundingToolCall):
                tool_call = {
                    "id": t.id,
                    "type": t.type,
                    t.type: t.bing_grounding,
                }
            else:
                tool_details = t.as_dict()[t.type]

                tool_call = {
                    "id": t.id,
                    "type": t.type,
                    t.type: tool_details,
                }
            tool_calls.append(tool_call)
        return tool_calls

    def _add_tool_call_event(
        self,
        span,
        step: RunStep,
        event_name: str,
        is_run_step_listing: bool = False,
    ) -> None:
        """
        Adds a tool call event to a span.

        This method processes tool calls from a given run step and adds them as an event
        to the provided span. It includes relevant attributes such as the run step status,
        timestamps, tool call details, and optionally the message status.

        :param span: The span instance where the tool call event will be recorded.
        :type span: AbstractSpan
        :param step: The run step containing details about the tool calls to be processed.
        :type step: RunStep
        :param event_name: The name of the event to be added to the span (e.g., "gen_ai.run_step.tool_calls").
        :type event_name: str
        :param is_run_step_listing: A flag indicating whether the event is part of a run step listing.
            If True, the run step status is included in the attributes; otherwise, the message status is included.
        :type is_run_step_listing: bool
        :return: None
        """
        tool_calls = self._process_tool_calls(step)

        run_step_status = None
        message_status = None
        if is_run_step_listing:
            run_step_status = step.status
        else:
            message_status = step.status

        attributes = self._create_event_attributes(
            thread_id=step.thread_id,
            agent_id=step.agent_id,
            thread_run_id=step.run_id,
            message_status=message_status,
            run_step_status=run_step_status,
            created_at=step.created_at,
            completed_at=step.completed_at,
            cancelled_at=step.cancelled_at,
            failed_at=step.failed_at,
            run_step_last_error=step.last_error,
            usage=step.usage,
        )

        if tool_calls:
            attributes[GEN_AI_EVENT_CONTENT] = json.dumps({"tool_calls": tool_calls}, ensure_ascii=False)

        span.span_instance.add_event(name=event_name, attributes=attributes)

    def add_run_step_event(self, span, step: RunStep) -> None:
        """
        Adds a run step event to the span.

        This method determines the type of the run step and adds the appropriate event
        to the provided span. It processes either a "message_creation" or "tool_calls"
        run step and includes relevant attributes such as the run step status, timestamps,
        and tool call or message details.

        :param span: The span instance where the run step event will be recorded.
        :type span: AbstractSpan
        :param step: The run step containing details about the event to be added.
        :type step: RunStep
        :return: None
        """
        if step["type"] == "message_creation":
            self._add_message_creation_run_step_event(span, step)
        elif step["type"] == "tool_calls":
            self._add_tool_call_event(span, step, "gen_ai.run_step.tool_calls", is_run_step_listing=True)

    def _add_message_creation_run_step_event(self, span, step: RunStep) -> None:
        attributes = self._create_event_attributes(
            thread_id=step.thread_id,
            agent_id=step.agent_id,
            thread_run_id=step.run_id,
            message_id=step["step_details"]["message_creation"]["message_id"],
            run_step_status=step.status,
            created_at=step.created_at,
            completed_at=step.completed_at,
            cancelled_at=step.cancelled_at,
            failed_at=step.failed_at,
            run_step_last_error=step.last_error,
            usage=step.usage,
        )
        span.span_instance.add_event(name="gen_ai.run_step.message_creation", attributes=attributes)

    def _add_message_event(
        self,
        span,
        role: str,
        content: Any,
        attachments: Any = None,  # Optional[List[MessageAttachment]] or dict
        thread_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        message_id: Optional[str] = None,
        thread_run_id: Optional[str] = None,
        message_status: Optional[str] = None,
        incomplete_details: Optional[MessageIncompleteDetails] = None,
        usage: Optional[_models.RunStepCompletionUsage] = None,
    ) -> None:
        # TODO document new fields

        event_body = {}
        if _trace_agents_content:
            event_body["content"] = content
            if attachments:
                event_body["attachments"] = []
                for attachment in attachments:
                    attachment_body = {"id": attachment.file_id}
                    if attachment.tools:
                        attachment_body["tools"] = [self._get_field(tool, "type") for tool in attachment.tools]
                    event_body["attachments"].append(attachment_body)

        if incomplete_details:
            event_body["incomplete_details"] = incomplete_details
        event_body["role"] = role

        attributes = self._create_event_attributes(
            thread_id=thread_id,
            agent_id=agent_id,
            thread_run_id=thread_run_id,
            message_id=message_id,
            message_status=message_status,
            usage=usage,
        )
        attributes[GEN_AI_EVENT_CONTENT] = json.dumps(event_body, ensure_ascii=False)
        span.span_instance.add_event(name=f"gen_ai.{role}.message", attributes=attributes)

    def _get_field(self, obj: Any, field: str) -> Any:
        if not obj:
            return None

        if isinstance(obj, dict):
            return obj.get(field, None)

        return getattr(obj, field, None)

    def _add_instructions_event(
        self,
        span: "AbstractSpan",
        instructions: Optional[str],
        additional_instructions: Optional[str],
        agent_id: Optional[str] = None,
        thread_id: Optional[str] = None,
    ) -> None:
        if not instructions:
            return

        event_body: Dict[str, Any] = {}
        if _trace_agents_content and (instructions or additional_instructions):
            if instructions and additional_instructions:
                event_body["content"] = f"{instructions} {additional_instructions}"
            else:
                event_body["content"] = instructions or additional_instructions

        attributes = self._create_event_attributes(agent_id=agent_id, thread_id=thread_id)
        attributes[GEN_AI_EVENT_CONTENT] = json.dumps(event_body, ensure_ascii=False)
        span.span_instance.add_event(name=GEN_AI_SYSTEM_MESSAGE, attributes=attributes)

    def _get_role(self, role: Optional[Union[str, MessageRole]]) -> str:
        if role is None or role is _Unset:
            return "user"

        if isinstance(role, MessageRole):
            return role.value

        return role

    def _status_to_string(self, status: Any) -> str:
        return status.value if hasattr(status, "value") else status

    def _add_tool_assistant_message_event(self, span, step: RunStep) -> None:
        self._add_tool_call_event(span, step, "gen_ai.assistant.message", is_run_step_listing=False)

    def set_end_run(self, span: "AbstractSpan", run: Optional[ThreadRun]) -> None:
        if run and span and span.span_instance.is_recording:
            span.add_attribute(GEN_AI_THREAD_RUN_STATUS, self._status_to_string(run.status))
            span.add_attribute(GEN_AI_RESPONSE_MODEL, run.model)
            if run and run.usage:
                span.add_attribute(GEN_AI_USAGE_INPUT_TOKENS, run.usage.prompt_tokens)
                span.add_attribute(GEN_AI_USAGE_OUTPUT_TOKENS, run.usage.completion_tokens)

    @staticmethod
    def agent_api_response_to_str(response_format: Any) -> Optional[str]:
        """
        Convert response_format to string.

        :param response_format: The response format.
        :type response_format: ~azure.ai.projects._types.AgentsApiResponseFormatOption
        :returns: string for the response_format.
        :rtype: Optional[str]
        :raises: Value error if response_format is not of type AgentsApiResponseFormatOption.
        """
        if isinstance(response_format, str) or response_format is None:
            return response_format
        if isinstance(response_format, AgentsApiResponseFormatMode):
            return response_format.value
        if isinstance(response_format, _models.AgentsApiResponseFormat):
            return response_format.type
        if isinstance(response_format, _models.ResponseFormatJsonSchemaType):
            return response_format.type
        raise ValueError(f"Unknown response format {type(response_format)}")

    def start_thread_run_span(
        self,
        operation_name: OperationName,
        project_name: str,
        thread_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        model: Optional[str] = None,
        instructions: Optional[str] = None,
        additional_instructions: Optional[str] = None,
        additional_messages: Optional[List[ThreadMessage]] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        _tools: Optional[List[ToolDefinition]] = None,
        max_prompt_tokens: Optional[int] = None,
        max_completion_tokens: Optional[int] = None,
        response_format: Optional["_types.AgentsApiResponseFormatOption"] = None,
    ) -> "Optional[AbstractSpan]":
        span = start_span(
            operation_name,
            project_name,
            thread_id=thread_id,
            agent_id=agent_id,
            model=model,
            temperature=temperature,
            top_p=top_p,
            max_prompt_tokens=max_prompt_tokens,
            max_completion_tokens=max_completion_tokens,
            response_format=_AIAgentsInstrumentorPreview.agent_api_response_to_str(response_format),
        )
        if span and span.span_instance.is_recording and instructions and additional_instructions:
            self._add_instructions_event(
                span, instructions, additional_instructions, thread_id=thread_id, agent_id=agent_id
            )

            if additional_messages:
                for message in additional_messages:
                    self.add_thread_message_event(span, message)
        return span

    def start_submit_tool_outputs_span(
        self,
        project_name: str,
        thread_id: Optional[str] = None,
        run_id: Optional[str] = None,
        tool_outputs: Optional[List[ToolOutput]] = None,
        event_handler: Optional[Union[AgentEventHandler, AsyncAgentEventHandler]] = None,
    ) -> "Optional[AbstractSpan]":
        run_span = event_handler.span if isinstance(event_handler, _AgentEventHandlerTraceWrapper) else None
        if run_span is None:
            run_span = event_handler.span if isinstance(event_handler, _AsyncAgentEventHandlerTraceWrapper) else None

        if run_span:
            recorded = self._add_tool_message_events(run_span, tool_outputs)
        else:
            recorded = False

        span = start_span(OperationName.SUBMIT_TOOL_OUTPUTS, project_name, thread_id=thread_id, run_id=run_id)
        if not recorded:
            self._add_tool_message_events(span, tool_outputs)
        return span

    def _add_tool_message_events(
        self, span: "Optional[AbstractSpan]", tool_outputs: Optional[List[ToolOutput]]
    ) -> bool:
        if span and span.span_instance.is_recording and tool_outputs:
            for tool_output in tool_outputs:
                if _trace_agents_content:
                    body = {"content": tool_output["output"], "id": tool_output["tool_call_id"]}
                else:
                    body = {"content": "", "id": tool_output["tool_call_id"]}
                span.span_instance.add_event(
                    "gen_ai.tool.message", {"gen_ai.event.content": json.dumps(body, ensure_ascii=False)}
                )
            return True

        return False

    def start_create_agent_span(
        self,
        project_name: str,
        model: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        instructions: Optional[str] = None,
        _tools: Optional[List[ToolDefinition]] = None,
        _tool_resources: Optional[ToolResources] = None,
        _toolset: Optional[ToolSet] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        response_format: Optional["_types.AgentsApiResponseFormatOption"] = None,
    ) -> "Optional[AbstractSpan]":
        span = start_span(
            OperationName.CREATE_AGENT,
            project_name,
            span_name=f"{OperationName.CREATE_AGENT.value} {name}",
            model=model,
            temperature=temperature,
            top_p=top_p,
            response_format=_AIAgentsInstrumentorPreview.agent_api_response_to_str(response_format),
        )
        if span and span.span_instance.is_recording:
            if name:
                span.add_attribute(GEN_AI_AGENT_NAME, name)
            if description:
                span.add_attribute(GEN_AI_AGENT_DESCRIPTION, description)
            self._add_instructions_event(span, instructions, None)

        return span

    def start_create_thread_span(
        self,
        project_name: str,
        messages: Optional[List[ThreadMessage]] = None,
        _tool_resources: Optional[ToolResources] = None,
    ) -> "Optional[AbstractSpan]":
        span = start_span(OperationName.CREATE_THREAD, project_name)
        if span and span.span_instance.is_recording:
            for message in messages or []:
                self.add_thread_message_event(span, message)

        return span

    def start_list_messages_span(self, project_name: str, thread_id: Optional[str] = None) -> "Optional[AbstractSpan]":
        return start_span(OperationName.LIST_MESSAGES, project_name, thread_id=thread_id)

    def start_list_run_steps_span(
        self, project_name: str, run_id: Optional[str] = None, thread_id: Optional[str] = None
    ) -> "Optional[AbstractSpan]":
        return start_span(OperationName.LIST_RUN_STEPS, project_name, run_id=run_id, thread_id=thread_id)

    def trace_create_agent(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        name = kwargs.get("name")
        model = kwargs.get("model")
        description = kwargs.get("description")
        instructions = kwargs.get("instructions")
        tools = kwargs.get("tools")
        tool_resources = kwargs.get("tool_resources")
        toolset = kwargs.get("toolset")
        temperature = kwargs.get("temperature")
        top_p = kwargs.get("top_p")
        response_format = kwargs.get("response_format")

        span = self.start_create_agent_span(
            project_name=project_name,
            name=name,
            model=model,
            description=description,
            instructions=instructions,
            _tools=tools,
            _tool_resources=tool_resources,
            _toolset=toolset,
            temperature=temperature,
            top_p=top_p,
            response_format=response_format,
        )

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                span.add_attribute(GEN_AI_AGENT_ID, result.id)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_create_agent_async(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        name = kwargs.get("name")
        model = kwargs.get("model")
        description = kwargs.get("description")
        instructions = kwargs.get("instructions")
        tools = kwargs.get("tools")
        tool_resources = kwargs.get("tool_resources")
        toolset = kwargs.get("toolset")
        temperature = kwargs.get("temperature")
        top_p = kwargs.get("top_p")
        response_format = kwargs.get("response_format")

        span = self.start_create_agent_span(
            project_name=project_name,
            name=name,
            model=model,
            description=description,
            instructions=instructions,
            _tools=tools,
            _tool_resources=tool_resources,
            _toolset=toolset,
            temperature=temperature,
            top_p=top_p,
            response_format=response_format,
        )

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                span.add_attribute(GEN_AI_AGENT_ID, result.id)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_create_thread(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        messages = kwargs.get("messages")

        span = self.start_create_thread_span(project_name=project_name, messages=messages)

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                span.add_attribute(GEN_AI_THREAD_ID, result.get("id"))
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_create_thread_async(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        messages = kwargs.get("messages")

        span = self.start_create_thread_span(project_name=project_name, messages=messages)

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                span.add_attribute(GEN_AI_THREAD_ID, result.get("id"))
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_create_message(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        role = kwargs.get("role")
        content = kwargs.get("content")
        attachments = kwargs.get("attachments")

        span = self.start_create_message_span(
            project_name=project_name, thread_id=thread_id, content=content, role=role, attachments=attachments
        )

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                span.add_attribute(GEN_AI_MESSAGE_ID, result.get("id"))
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_create_message_async(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        role = kwargs.get("role")
        content = kwargs.get("content")
        attachments = kwargs.get("attachments")

        span = self.start_create_message_span(
            project_name=project_name, thread_id=thread_id, content=content, role=role, attachments=attachments
        )

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                span.add_attribute(GEN_AI_MESSAGE_ID, result.get("id"))
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_create_run(self, operation_name, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        agent_id = kwargs.get("agent_id")
        model = kwargs.get("model")
        instructions = kwargs.get("instructions")
        additional_instructions = kwargs.get("additional_instructions")
        additional_messages = kwargs.get("additional_messages")
        temperature = kwargs.get("temperature")
        tools = kwargs.get("tools")
        top_p = kwargs.get("top_p")
        max_prompt_tokens = kwargs.get("max_prompt_tokens")
        max_completion_tokens = kwargs.get("max_completion_tokens")
        response_format = kwargs.get("response_format")

        span = self.start_thread_run_span(
            operation_name,
            project_name,
            thread_id,
            agent_id,
            model=model,
            instructions=instructions,
            additional_instructions=additional_instructions,
            additional_messages=additional_messages,
            temperature=temperature,
            _tools=tools,
            top_p=top_p,
            max_prompt_tokens=max_prompt_tokens,
            max_completion_tokens=max_completion_tokens,
            response_format=response_format,
        )

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                self.set_end_run(span, result)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_create_run_async(self, operation_name, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        agent_id = kwargs.get("agent_id")
        model = kwargs.get("model")
        instructions = kwargs.get("instructions")
        additional_instructions = kwargs.get("additional_instructions")
        additional_messages = kwargs.get("additional_messages")
        temperature = kwargs.get("temperature")
        tools = kwargs.get("tools")
        top_p = kwargs.get("top_p")
        max_prompt_tokens = kwargs.get("max_prompt_tokens")
        max_completion_tokens = kwargs.get("max_completion_tokens")
        response_format = kwargs.get("response_format")

        span = self.start_thread_run_span(
            operation_name,
            project_name,
            thread_id,
            agent_id,
            model=model,
            instructions=instructions,
            additional_instructions=additional_instructions,
            additional_messages=additional_messages,
            temperature=temperature,
            _tools=tools,
            top_p=top_p,
            max_prompt_tokens=max_prompt_tokens,
            max_completion_tokens=max_completion_tokens,
            response_format=response_format,
        )

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                if span.span_instance.is_recording:
                    span.add_attribute(GEN_AI_THREAD_RUN_STATUS, self._status_to_string(result.status))
                    span.add_attribute(GEN_AI_RESPONSE_MODEL, result.model)
                    if result.usage:
                        span.add_attribute(GEN_AI_USAGE_INPUT_TOKENS, result.usage.prompt_tokens)
                        span.add_attribute(GEN_AI_USAGE_OUTPUT_TOKENS, result.usage.completion_tokens)
                        span.add_attribute(GEN_AI_MESSAGE_ID, result.get("id"))
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_submit_tool_outputs(self, stream, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        run_id = kwargs.get("run_id")
        tool_outputs = kwargs.get("tool_outputs")
        event_handler = kwargs.get("event_handler")

        span = self.start_submit_tool_outputs_span(
            project_name=project_name,
            thread_id=thread_id,
            run_id=run_id,
            tool_outputs=tool_outputs,
            event_handler=event_handler,
        )

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                if stream and event_handler:
                    kwargs["event_handler"] = self.wrap_handler(event_handler, span)

                result = function(*args, **kwargs)
                if not isinstance(result, AgentRunStream):
                    self.set_end_run(span, result)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_submit_tool_outputs_async(self, stream, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        run_id = kwargs.get("run_id")
        tool_outputs = kwargs.get("tool_outputs")
        event_handler = kwargs.get("event_handler")

        span = self.start_submit_tool_outputs_span(
            project_name=project_name,
            thread_id=thread_id,
            run_id=run_id,
            tool_outputs=tool_outputs,
            event_handler=event_handler,
        )

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                if stream:
                    kwargs["event_handler"] = self.wrap_async_handler(event_handler, span)

                result = await function(*args, **kwargs)
                if not isinstance(result, AsyncAgentRunStream):
                    self.set_end_run(span, result)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_handle_submit_tool_outputs(self, function, *args, **kwargs):
        event_handler = kwargs.get("event_handler")
        if event_handler is None:
            event_handler = args[2]
        span = getattr(event_handler, "span", None)

        if span is None:
            return function(*args, **kwargs)

        with span.change_context(span.span_instance):
            try:
                result = function(*args, **kwargs)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_handle_submit_tool_outputs_async(self, function, *args, **kwargs):
        event_handler = kwargs.get("event_handler")
        if event_handler is None:
            event_handler = args[2]
        span = getattr(event_handler, "span", None)

        if span is None:
            return await function(*args, **kwargs)

        with span.change_context(span.span_instance):
            try:
                result = await function(*args, **kwargs)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_create_stream(self, function, *args, **kwargs):
        operation_name = OperationName.PROCESS_THREAD_RUN
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        agent_id = kwargs.get("agent_id")
        model = kwargs.get("model")
        instructions = kwargs.get("instructions")
        additional_instructions = kwargs.get("additional_instructions")
        additional_messages = kwargs.get("additional_messages")
        temperature = kwargs.get("temperature")
        tools = kwargs.get("tools")
        top_p = kwargs.get("top_p")
        max_prompt_tokens = kwargs.get("max_prompt_tokens")
        max_completion_tokens = kwargs.get("max_completion_tokens")
        response_format = kwargs.get("response_format")
        event_handler = kwargs.get("event_handler")

        span = self.start_thread_run_span(
            operation_name,
            project_name,
            thread_id,
            agent_id,
            model=model,
            instructions=instructions,
            additional_instructions=additional_instructions,
            additional_messages=additional_messages,
            temperature=temperature,
            _tools=tools,
            top_p=top_p,
            max_prompt_tokens=max_prompt_tokens,
            max_completion_tokens=max_completion_tokens,
            response_format=response_format,
        )

        if span is None:
            return function(*args, **kwargs)

        with span.change_context(span.span_instance):
            try:
                kwargs["event_handler"] = self.wrap_handler(event_handler, span)
                result = function(*args, **kwargs)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_create_stream_async(self, function, *args, **kwargs):
        operation_name = OperationName.PROCESS_THREAD_RUN
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")
        agent_id = kwargs.get("agent_id")
        model = kwargs.get("model")
        instructions = kwargs.get("instructions")
        additional_instructions = kwargs.get("additional_instructions")
        additional_messages = kwargs.get("additional_messages")
        temperature = kwargs.get("temperature")
        tools = kwargs.get("tools")
        top_p = kwargs.get("top_p")
        max_prompt_tokens = kwargs.get("max_prompt_tokens")
        max_completion_tokens = kwargs.get("max_completion_tokens")
        response_format = kwargs.get("response_format")
        event_handler = kwargs.get("event_handler")

        span = self.start_thread_run_span(
            operation_name,
            project_name,
            thread_id,
            agent_id,
            model=model,
            instructions=instructions,
            additional_instructions=additional_instructions,
            additional_messages=additional_messages,
            temperature=temperature,
            _tools=tools,
            top_p=top_p,
            max_prompt_tokens=max_prompt_tokens,
            max_completion_tokens=max_completion_tokens,
            response_format=response_format,
        )

        if span is None:
            return await function(*args, **kwargs)

        # TODO: how to keep span active in the current context without existing?
        # TODO: dummy span for none
        with span.change_context(span.span_instance):
            try:
                kwargs["event_handler"] = self.wrap_async_handler(event_handler, span)
                result = await function(*args, **kwargs)
            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_list_messages(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")

        span = self.start_list_messages_span(project_name=project_name, thread_id=thread_id)

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                for message in result.data:
                    self.add_thread_message_event(span, message)

            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def trace_list_run_steps(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        run_id = kwargs.get("run_id")
        thread_id = kwargs.get("thread_id")

        span = self.start_list_run_steps_span(project_name=project_name, run_id=run_id, thread_id=thread_id)

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = function(*args, **kwargs)
                if hasattr(result, "data") and result.data is not None:
                    for step in result.data:
                        self.add_run_step_event(span, step)

            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_list_run_steps_async(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        run_id = kwargs.get("run_id")
        thread_id = kwargs.get("thread_id")

        span = self.start_list_run_steps_span(project_name=project_name, run_id=run_id, thread_id=thread_id)

        if span is None:
            return function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                if hasattr(result, "data") and result.data is not None:
                    for step in result.data:
                        self.add_run_step_event(span, step)

            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    async def trace_list_messages_async(self, function, *args, **kwargs):
        project_name = args[  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
            0
        ]._config.project_name
        thread_id = kwargs.get("thread_id")

        span = self.start_list_messages_span(project_name=project_name, thread_id=thread_id)

        if span is None:
            return await function(*args, **kwargs)

        with span:
            try:
                result = await function(*args, **kwargs)
                for message in result.data:
                    self.add_thread_message_event(span, message)

            except Exception as exc:
                # Set the span status to error
                if isinstance(span.span_instance, Span):  # pyright: ignore [reportPossiblyUnboundVariable]
                    span.span_instance.set_status(
                        StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                        description=str(exc),
                    )
                module = getattr(exc, "__module__", "")
                module = module if module != "builtins" else ""
                error_type = f"{module}.{type(exc).__name__}" if module else type(exc).__name__
                self._set_attributes(span, ("error.type", error_type))
                raise

        return result

    def handle_run_stream_exit(self, _function, *args, **kwargs):
        agent_run_stream = args[0]
        exc_type = kwargs.get("exc_type")
        exc_val = kwargs.get("exc_val")
        exc_tb = kwargs.get("exc_tb")
        # TODO: is it a good idea?
        # if not, we'll need to wrap stream and call exit
        if (
            agent_run_stream.event_handler
            and agent_run_stream.event_handler.__class__.__name__ == "_AgentEventHandlerTraceWrapper"
        ):
            agent_run_stream.event_handler.__exit__(exc_type, exc_val, exc_tb)
        elif (
            agent_run_stream.event_handler
            and agent_run_stream.event_handler.__class__.__name__ == "_AsyncAgentEventHandlerTraceWrapper"
        ):
            agent_run_stream.event_handler.__aexit__(exc_type, exc_val, exc_tb)

    def wrap_handler(
        self, handler: "Optional[AgentEventHandler]" = None, span: "Optional[AbstractSpan]" = None
    ) -> "Optional[AgentEventHandler]":
        # Do not create a handler wrapper if we do not have handler in the first place.
        if not handler:
            return None

        if isinstance(handler, _AgentEventHandlerTraceWrapper):
            return handler

        if span and span.span_instance.is_recording:
            return _AgentEventHandlerTraceWrapper(self, span, handler)

        return handler

    def wrap_async_handler(
        self, handler: "Optional[AsyncAgentEventHandler]" = None, span: "Optional[AbstractSpan]" = None
    ) -> "Optional[AsyncAgentEventHandler]":
        # Do not create a handler wrapper if we do not have handler in the first place.
        if not handler:
            return None

        if isinstance(handler, _AsyncAgentEventHandlerTraceWrapper):
            return handler

        if span and span.span_instance.is_recording:
            return _AsyncAgentEventHandlerTraceWrapper(self, span, handler)

        return handler

    def start_create_message_span(
        self,
        project_name: str,
        thread_id: Optional[str] = None,
        content: Optional[str] = None,
        role: Optional[Union[str, MessageRole]] = None,
        attachments: Optional[List[MessageAttachment]] = None,
    ) -> "Optional[AbstractSpan]":
        role_str = self._get_role(role)
        span = start_span(OperationName.CREATE_MESSAGE, project_name, thread_id=thread_id)
        if span and span.span_instance.is_recording:
            self._add_message_event(span, role_str, content, attachments=attachments, thread_id=thread_id)
        return span

    def _trace_sync_function(
        self,
        function: Callable,
        *,
        _args_to_ignore: Optional[List[str]] = None,
        _trace_type=TraceType.AGENTS,
        _name: Optional[str] = None,
    ) -> Callable:
        """
        Decorator that adds tracing to a synchronous function.

        :param function: The function to be traced.
        :type function: Callable
        :param args_to_ignore: A list of argument names to be ignored in the trace.
                            Defaults to None.
        :type: args_to_ignore: [List[str]], optional
        :param trace_type: The type of the trace. Defaults to TraceType.AGENTS.
        :type trace_type: TraceType, optional
        :param name: The name of the trace, will set to func name if not provided.
        :type name: str, optional
        :return: The traced function.
        :rtype: Callable
        """

        @functools.wraps(function)
        def inner(*args, **kwargs):  # pylint: disable=R0911
            span_impl_type = settings.tracing_implementation()  # pylint: disable=E1102
            if span_impl_type is None:
                return function(*args, **kwargs)

            class_function_name = function.__qualname__

            if class_function_name.startswith("AgentsOperations.create_agent"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_agent(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_thread"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_thread(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_message"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_message(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_run"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_run(OperationName.START_THREAD_RUN, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_and_process_run"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_run(OperationName.PROCESS_THREAD_RUN, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.submit_tool_outputs_to_run"):
                kwargs.setdefault("merge_span", True)
                return self.trace_submit_tool_outputs(False, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.submit_tool_outputs_to_stream"):
                kwargs.setdefault("merge_span", True)
                return self.trace_submit_tool_outputs(True, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations._handle_submit_tool_outputs"):
                return self.trace_handle_submit_tool_outputs(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_stream"):
                kwargs.setdefault("merge_span", True)
                return self.trace_create_stream(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.list_messages"):
                kwargs.setdefault("merge_span", True)
                return self.trace_list_messages(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.list_run_steps"):
                return self.trace_list_run_steps(function, *args, **kwargs)
            if class_function_name.startswith("AgentRunStream.__exit__"):
                return self.handle_run_stream_exit(function, *args, **kwargs)
            # Handle the default case (if the function name does not match)
            return None  # Ensure all paths return

        return inner

    def _trace_async_function(
        self,
        function: Callable,
        *,
        _args_to_ignore: Optional[List[str]] = None,
        _trace_type=TraceType.AGENTS,
        _name: Optional[str] = None,
    ) -> Callable:
        """
        Decorator that adds tracing to an asynchronous function.

        :param function: The function to be traced.
        :type function: Callable
        :param args_to_ignore: A list of argument names to be ignored in the trace.
                            Defaults to None.
        :type: args_to_ignore: [List[str]], optional
        :param trace_type: The type of the trace. Defaults to TraceType.AGENTS.
        :type trace_type: TraceType, optional
        :param name: The name of the trace, will set to func name if not provided.
        :type name: str, optional
        :return: The traced function.
        :rtype: Callable
        """

        @functools.wraps(function)
        async def inner(*args, **kwargs):  # pylint: disable=R0911
            span_impl_type = settings.tracing_implementation()  # pylint: disable=E1102
            if span_impl_type is None:
                return function(*args, **kwargs)

            class_function_name = function.__qualname__

            if class_function_name.startswith("AgentsOperations.create_agent"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_agent_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_thread"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_thread_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_message"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_message_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_run"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_run_async(OperationName.START_THREAD_RUN, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_and_process_run"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_run_async(OperationName.PROCESS_THREAD_RUN, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.submit_tool_outputs_to_run"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_submit_tool_outputs_async(False, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.submit_tool_outputs_to_stream"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_submit_tool_outputs_async(True, function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations._handle_submit_tool_outputs"):
                return await self.trace_handle_submit_tool_outputs_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.create_stream"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_create_stream_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.list_messages"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_list_messages_async(function, *args, **kwargs)
            if class_function_name.startswith("AgentsOperations.list_run_steps"):
                kwargs.setdefault("merge_span", True)
                return await self.trace_list_run_steps_async(function, *args, **kwargs)
            if class_function_name.startswith("AsyncAgentRunStream.__aexit__"):
                return self.handle_run_stream_exit(function, *args, **kwargs)
            # Handle the default case (if the function name does not match)
            return None  # Ensure all paths return

        return inner

    def _inject_async(self, f, _trace_type, _name):
        wrapper_fun = self._trace_async_function(f)
        wrapper_fun._original = f  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
        return wrapper_fun

    def _inject_sync(self, f, _trace_type, _name):
        wrapper_fun = self._trace_sync_function(f)
        wrapper_fun._original = f  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
        return wrapper_fun

    def _agents_apis(self):
        sync_apis = (
            ("azure.ai.projects.operations", "AgentsOperations", "create_agent", TraceType.AGENTS, "agent_create"),
            ("azure.ai.projects.operations", "AgentsOperations", "create_thread", TraceType.AGENTS, "thread_create"),
            ("azure.ai.projects.operations", "AgentsOperations", "create_message", TraceType.AGENTS, "message_create"),
            ("azure.ai.projects.operations", "AgentsOperations", "create_run", TraceType.AGENTS, "create_run"),
            (
                "azure.ai.projects.operations",
                "AgentsOperations",
                "create_and_process_run",
                TraceType.AGENTS,
                "create_and_process_run",
            ),
            (
                "azure.ai.projects.operations",
                "AgentsOperations",
                "submit_tool_outputs_to_run",
                TraceType.AGENTS,
                "submit_tool_outputs_to_run",
            ),
            (
                "azure.ai.projects.operations",
                "AgentsOperations",
                "submit_tool_outputs_to_stream",
                TraceType.AGENTS,
                "submit_tool_outputs_to_stream",
            ),
            (
                "azure.ai.projects.operations",
                "AgentsOperations",
                "_handle_submit_tool_outputs",
                TraceType.AGENTS,
                "_handle_submit_tool_outputs",
            ),
            ("azure.ai.projects.operations", "AgentsOperations", "create_stream", TraceType.AGENTS, "create_stream"),
            ("azure.ai.projects.operations", "AgentsOperations", "list_messages", TraceType.AGENTS, "list_messages"),
            ("azure.ai.projects.operations", "AgentsOperations", "list_run_steps", TraceType.AGENTS, "list_run_steps"),
            ("azure.ai.projects.models", "AgentRunStream", "__exit__", TraceType.AGENTS, "__exit__"),
        )
        async_apis = (
            ("azure.ai.projects.aio.operations", "AgentsOperations", "create_agent", TraceType.AGENTS, "agent_create"),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "create_thread",
                TraceType.AGENTS,
                "agents_thread_create",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "create_message",
                TraceType.AGENTS,
                "agents_thread_message",
            ),
            ("azure.ai.projects.aio.operations", "AgentsOperations", "create_run", TraceType.AGENTS, "create_run"),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "create_and_process_run",
                TraceType.AGENTS,
                "create_and_process_run",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "submit_tool_outputs_to_run",
                TraceType.AGENTS,
                "submit_tool_outputs_to_run",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "submit_tool_outputs_to_stream",
                TraceType.AGENTS,
                "submit_tool_outputs_to_stream",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "_handle_submit_tool_outputs",
                TraceType.AGENTS,
                "_handle_submit_tool_outputs",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "create_stream",
                TraceType.AGENTS,
                "create_stream",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "list_messages",
                TraceType.AGENTS,
                "list_messages",
            ),
            (
                "azure.ai.projects.aio.operations",
                "AgentsOperations",
                "list_run_steps",
                TraceType.AGENTS,
                "list_run_steps",
            ),
            ("azure.ai.projects.models", "AsyncAgentRunStream", "__aexit__", TraceType.AGENTS, "__aexit__"),
        )
        return sync_apis, async_apis

    def _agents_api_list(self):
        sync_apis, async_apis = self._agents_apis()
        yield sync_apis, self._inject_sync
        yield async_apis, self._inject_async

    def _generate_api_and_injector(self, apis):
        for api, injector in apis:
            for module_name, class_name, method_name, trace_type, name in api:
                try:
                    module = importlib.import_module(module_name)
                    api = getattr(module, class_name)
                    if hasattr(api, method_name):
                        yield api, method_name, trace_type, injector, name
                except AttributeError as e:
                    # Log the attribute exception with the missing class information
                    logging.warning(
                        "AttributeError: The module '%s' does not have the class '%s'. %s",
                        module_name,
                        class_name,
                        str(e),
                    )
                except Exception as e:  # pylint: disable=broad-except
                    # Log other exceptions as a warning, as we are not sure what they might be
                    logging.warning("An unexpected error occurred: '%s'", str(e))

    def _available_agents_apis_and_injectors(self):
        """
        Generates a sequence of tuples containing Agents API classes, method names, and
        corresponding injector functions.

        :return: A generator yielding tuples.
        :rtype: tuple
        """
        yield from self._generate_api_and_injector(self._agents_api_list())

    def _instrument_agents(self, enable_content_tracing: bool = False):
        """This function modifies the methods of the Agents API classes to
        inject logic before calling the original methods.
        The original methods are stored as _original attributes of the methods.

        :param enable_content_tracing: Indicates whether tracing of message content should be enabled.
                                    This also controls whether function call tool function names,
                                    parameter names and parameter values are traced.
        :type enable_content_tracing: bool
        """
        # pylint: disable=W0603
        global _agents_traces_enabled
        global _trace_agents_content
        if _agents_traces_enabled:
            raise RuntimeError("Traces already started for AI Agents")
        _agents_traces_enabled = True
        _trace_agents_content = enable_content_tracing
        for (
            api,
            method,
            trace_type,
            injector,
            name,
        ) in self._available_agents_apis_and_injectors():
            # Check if the method of the api class has already been modified
            if not hasattr(getattr(api, method), "_original"):
                setattr(api, method, injector(getattr(api, method), trace_type, name))

    def _uninstrument_agents(self):
        """This function restores the original methods of the Agents API classes
        by assigning them back from the _original attributes of the modified methods.
        """
        # pylint: disable=W0603
        global _agents_traces_enabled
        global _trace_agents_content
        _trace_agents_content = False
        for api, method, _, _, _ in self._available_agents_apis_and_injectors():
            if hasattr(getattr(api, method), "_original"):
                setattr(api, method, getattr(getattr(api, method), "_original"))
        _agents_traces_enabled = False

    def _is_instrumented(self):
        """This function returns True if Agents API has already been instrumented
        for tracing and False if it has not been instrumented.

        :return: A value indicating whether the Agents API is currently instrumented or not.
        :rtype: bool
        """
        return _agents_traces_enabled

    def _set_enable_content_recording(self, enable_content_recording: bool = False) -> None:
        """This function sets the content recording value.

        :param enable_content_recording: Indicates whether tracing of message content should be enabled.
                                    This also controls whether function call tool function names,
                                    parameter names and parameter values are traced.
        :type enable_content_recording: bool
        """
        global _trace_agents_content  # pylint: disable=W0603
        _trace_agents_content = enable_content_recording

    def _is_content_recording_enabled(self) -> bool:
        """This function gets the content recording value.

        :return: A bool value indicating whether content tracing is enabled.
        :rtype bool
        """
        return _trace_agents_content


class _AgentEventHandlerTraceWrapper(AgentEventHandler):
    def __init__(
        self,
        instrumentor: _AIAgentsInstrumentorPreview,
        span: "AbstractSpan",
        inner_handler: Optional[AgentEventHandler] = None,
    ):
        super().__init__()
        self.span = span
        self.inner_handler = inner_handler
        self.ended = False
        self.last_run: Optional[ThreadRun] = None
        self.last_message: Optional[ThreadMessage] = None
        self.instrumentor = instrumentor

    def initialize(
        self,
        response_iterator,
        submit_tool_outputs,
    ) -> None:
        self.submit_tool_outputs = submit_tool_outputs
        if self.inner_handler:
            self.inner_handler.initialize(response_iterator=response_iterator, submit_tool_outputs=submit_tool_outputs)

    def __next__(self) -> Any:
        if self.inner_handler:
            event_bytes = self.inner_handler.__next_impl__()
            return self._process_event(event_bytes.decode("utf-8"))
        return None

    # pylint: disable=R1710
    def on_message_delta(self, delta: "MessageDeltaChunk") -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return self.inner_handler.on_message_delta(delta)  # type: ignore

    def on_thread_message(self, message: "ThreadMessage") -> None:  # type: ignore[func-returns-value]
        retval = None
        if self.inner_handler:
            retval = self.inner_handler.on_thread_message(message)  # type: ignore

        if message.status in {"completed", "incomplete"}:
            self.last_message = message

        return retval  # type: ignore

    def on_thread_run(self, run: "ThreadRun") -> None:  # type: ignore[func-returns-value]
        retval = None

        if self.inner_handler:
            retval = self.inner_handler.on_thread_run(run)  # type: ignore
        self.last_run = run

        return retval  # type: ignore

    def on_run_step(self, step: "RunStep") -> None:  # type: ignore[func-returns-value]
        retval = None
        if self.inner_handler:
            retval = self.inner_handler.on_run_step(step)  # type: ignore

        if (
            step.type == "tool_calls"
            and isinstance(step.step_details, RunStepToolCallDetails)
            and step.status == RunStepStatus.COMPLETED
        ):
            self.instrumentor._add_tool_assistant_message_event(  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
                self.span, step
            )
        elif step.type == "message_creation" and step.status == RunStepStatus.COMPLETED:
            self.instrumentor.add_thread_message_event(self.span, cast(ThreadMessage, self.last_message), step.usage)
            self.last_message = None

        return retval  # type: ignore

    def on_run_step_delta(self, delta: "RunStepDeltaChunk") -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return self.inner_handler.on_run_step_delta(delta)  # type: ignore

    def on_error(self, data: str) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return self.inner_handler.on_error(data)  # type: ignore

    def on_done(self) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return self.inner_handler.on_done()  # type: ignore
        # it could be called multiple tines (for each step) __exit__

    def on_unhandled_event(self, event_type: str, event_data: Any) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return self.inner_handler.on_unhandled_event(event_type, event_data)  # type: ignore

    # pylint: enable=R1710

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.ended:
            self.ended = True
            self.instrumentor.set_end_run(self.span, self.last_run)

            if self.last_run and self.last_run.last_error:
                self.span.span_instance.set_status(
                    StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                    self.last_run.last_error.message,
                )
                self.span.add_attribute(ERROR_TYPE, self.last_run.last_error.code)

            self.span.__exit__(exc_type, exc_val, exc_tb)
            self.span.finish()


class _AsyncAgentEventHandlerTraceWrapper(AsyncAgentEventHandler):
    def __init__(
        self,
        instrumentor: _AIAgentsInstrumentorPreview,
        span: "AbstractSpan",
        inner_handler: Optional[AsyncAgentEventHandler] = None,
    ):
        super().__init__()
        self.span = span
        self.inner_handler = inner_handler
        self.ended = False
        self.last_run: Optional[ThreadRun] = None
        self.last_message: Optional[ThreadMessage] = None
        self.instrumentor = instrumentor

    def initialize(
        self,
        response_iterator,
        submit_tool_outputs,
    ) -> None:
        self.submit_tool_outputs = submit_tool_outputs
        if self.inner_handler:
            self.inner_handler.initialize(response_iterator=response_iterator, submit_tool_outputs=submit_tool_outputs)

    # cspell:disable-next-line
    async def __anext__(self) -> Any:
        if self.inner_handler:
            # cspell:disable-next-line
            event_bytes = await self.inner_handler.__anext_impl__()
            return await self._process_event(event_bytes.decode("utf-8"))

    # pylint: disable=R1710
    async def on_message_delta(self, delta: "MessageDeltaChunk") -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return await self.inner_handler.on_message_delta(delta)  # type: ignore

    async def on_thread_message(self, message: "ThreadMessage") -> None:  # type: ignore[func-returns-value]
        retval = None
        if self.inner_handler:
            retval = await self.inner_handler.on_thread_message(message)  # type: ignore

        if message.status in {"completed", "incomplete"}:
            self.last_message = message

        return retval  # type: ignore

    async def on_thread_run(self, run: "ThreadRun") -> None:  # type: ignore[func-returns-value]
        retval = None

        if self.inner_handler:
            retval = await self.inner_handler.on_thread_run(run)  # type: ignore
        self.last_run = run

        return retval  # type: ignore

    async def on_run_step(self, step: "RunStep") -> None:  # type: ignore[func-returns-value]
        retval = None
        if self.inner_handler:
            retval = await self.inner_handler.on_run_step(step)  # type: ignore

        if (
            step.type == "tool_calls"
            and isinstance(step.step_details, RunStepToolCallDetails)
            and step.status == RunStepStatus.COMPLETED
        ):
            self.instrumentor._add_tool_assistant_message_event(  # pylint: disable=protected-access # pyright: ignore [reportFunctionMemberAccess]
                self.span, step
            )
        elif step.type == "message_creation" and step.status == RunStepStatus.COMPLETED:
            self.instrumentor.add_thread_message_event(self.span, cast(ThreadMessage, self.last_message), step.usage)
            self.last_message = None

        return retval  # type: ignore

    async def on_run_step_delta(self, delta: "RunStepDeltaChunk") -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return await self.inner_handler.on_run_step_delta(delta)  # type: ignore

    async def on_error(self, data: str) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return await self.inner_handler.on_error(data)  # type: ignore

    async def on_done(self) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return await self.inner_handler.on_done()  # type: ignore
        # it could be called multiple tines (for each step) __exit__

    async def on_unhandled_event(self, event_type: str, event_data: Any) -> None:  # type: ignore[func-returns-value]
        if self.inner_handler:
            return await self.inner_handler.on_unhandled_event(event_type, event_data)  # type: ignore

    # pylint: enable=R1710

    def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.ended:
            self.ended = True
            self.instrumentor.set_end_run(self.span, self.last_run)

            if self.last_run and self.last_run.last_error:
                self.span.set_status(
                    StatusCode.ERROR,  # pyright: ignore [reportPossiblyUnboundVariable]
                    self.last_run.last_error.message,
                )
                self.span.add_attribute(ERROR_TYPE, self.last_run.last_error.code)

            self.span.__exit__(exc_type, exc_val, exc_tb)
            self.span.finish()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use agent operations with interation in streaming from
    the Azure Agents service using a asynchronous client.

USAGE:
    python sample_agents_stream_iteration_async.py

    Before running the sample:

    pip install azure-ai-projects azure-identity aiohttp

    Set this environment variables with your own values:
    PROJECT_CONNECTION_STRING - the Azure AI Project connection string, as found in your AI Foundry project.
"""
import asyncio

from azure.ai.projects.aio import AIProjectClient
from azure.ai.projects.models import AgentStreamEvent
from azure.ai.projects.models import MessageDeltaChunk, RunStep, ThreadMessage, ThreadRun
from azure.identity.aio import DefaultAzureCredential

import os


async def main() -> None:
    async with DefaultAzureCredential() as creds:
        async with AIProjectClient.from_connection_string(
            credential=creds, conn_str=os.environ["PROJECT_CONNECTION_STRING"]
        ) as project_client:
            agent = await project_client.agents.create_agent(
                model=os.environ["MODEL_DEPLOYMENT_NAME"], name="my-assistant", instructions="You are helpful assistant"
            )
            print(f"Created agent, agent ID: {agent.id}")

            thread = await project_client.agents.create_thread()
            print(f"Created thread, thread ID {thread.id}")

            message = await project_client.agents.create_message(
                thread_id=thread.id, role="user", content="Hello, tell me a joke"
            )
            print(f"Created message, message ID {message.id}")

            async with await project_client.agents.create_stream(thread_id=thread.id, agent_id=agent.id) as stream:
                async for event_type, event_data, _ in stream:

                    if isinstance(event_data, MessageDeltaChunk):
                        print(f"Text delta received: {event_data.text}")

                    elif isinstance(event_data, ThreadMessage):
                        print(f"ThreadMessage created. ID: {event_data.id}, Status: {event_data.status}")

                    elif isinstance(event_data, ThreadRun):
                        print(f"ThreadRun status: {event_data.status}")
                    elif isinstance(event_data, RunStep):
                        print(f"RunStep type: {event_data.type}, Status: {event_data.status}")

                    elif event_type == AgentStreamEvent.ERROR:
                        print(f"An error occurred. Data: {event_data}")

                    elif event_type == AgentStreamEvent.DONE:
                        print("Stream completed.")
                        break

                    else:
                        print(f"Unhandled Event Type: {event_type}, Data: {event_data}")

            await project_client.agents.delete_agent(agent.id)
            print("Deleted agent")

            messages = await project_client.agents.list_messages(thread_id=thread.id)
            print(f"Messages: {messages}")


if __name__ == "__main__":
    asyncio.run(main())



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use evaluation operations from
    the Azure Evaluation service using a asynchronous client.

USAGE:
    python sample_evaluation_async.py

    Before running the sample:

    pip install azure-ai-projects azure-identity aiohttp

    Set this environment variables with your own values:
    PROJECT_CONNECTION_STRING - the Azure AI Project connection string, as found in your AI Foundry project.
"""
import asyncio
import os

from azure.ai.projects.aio import AIProjectClient
from azure.identity.aio import DefaultAzureCredential
from azure.ai.projects.models import (
    Evaluation,
    Dataset,
    EvaluatorConfiguration,
    ConnectionType,
)
from azure.ai.evaluation import F1ScoreEvaluator, RelevanceEvaluator, ViolenceEvaluator


async def main():

    async with DefaultAzureCredential() as credential:

        async with AIProjectClient.from_connection_string(
            credential=credential,
            conn_str=os.environ["PROJECT_CONNECTION_STRING"],
        ) as project_client:

            # Upload data for evaluation
            data_id, _ = project_client.upload_file("./data/evaluate_test_data.jsonl")

            default_connection = await project_client.connections.get_default(
                connection_type=ConnectionType.AZURE_OPEN_AI
            )

            deployment_name = "<>"
            api_version = "<>"

            # Create an evaluation
            evaluation = Evaluation(
                display_name="Remote Evaluation",
                description="Evaluation of dataset",
                data=Dataset(id=data_id),
                evaluators={
                    "f1_score": EvaluatorConfiguration(
                        id=F1ScoreEvaluator.id,
                    ),
                    "relevance": EvaluatorConfiguration(
                        id=RelevanceEvaluator.id,
                        init_params={
                            "model_config": default_connection.to_evaluator_model_config(
                                deployment_name=deployment_name, api_version=api_version
                            )
                        },
                    ),
                    "violence": EvaluatorConfiguration(
                        id=ViolenceEvaluator.id,
                        init_params={"azure_ai_project": project_client.scope},
                    ),
                },
            )

            # Create evaluation
            evaluation_response = await project_client.evaluations.create(evaluation)

            # Get evaluation
            get_evaluation_response = await project_client.evaluations.get(evaluation_response.id)

            print("----------------------------------------------------------------")
            print("Created evaluation, evaluation ID: ", get_evaluation_response.id)
            print("Evaluation status: ", get_evaluation_response.status)
            if isinstance(get_evaluation_response.properties, dict):
                print(
                    "AI Foundry URI: ",
                    get_evaluation_response.properties["AiStudioEvaluationUri"],
                )
            print("----------------------------------------------------------------")


if __name__ == "__main__":
    asyncio.run(main())



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use agents with Logic Apps to execute the task of sending an email.
 
PREREQUISITES:
    1) Create a Logic App within the same resource group as your Azure AI Project in Azure Portal
    2) To configure your Logic App to send emails, you must include an HTTP request trigger that is 
    configured to accept JSON with 'to', 'subject', and 'body'. The guide to creating a Logic App Workflow
    can be found here: 
    https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/assistants-logic-apps#create-logic-apps-workflows-for-function-calling
    
USAGE:
    python sample_agents_logic_apps.py
 
    Before running the sample:
 
    pip install azure-ai-projects azure-identity

    Set this environment variables with your own values:
    1) PROJECT_CONNECTION_STRING - The project connection string, as found in the overview page of your
       Azure AI Foundry project.
    2) MODEL_DEPLOYMENT_NAME - The deployment name of the AI model, as found under the "Name" column in 
       the "Models + endpoints" tab in your Azure AI Foundry project.

    Replace the following values in the sample with your own values:
    1) <LOGIC_APP_NAME> - The name of the Logic App you created.
    2) <TRIGGER_NAME> - The name of the trigger in the Logic App you created (the default name for HTTP
        triggers in the Azure Portal is "When_a_HTTP_request_is_received").
    3) <RECIPIENT_EMAIL> - The email address of the recipient.
"""


import os
import requests
from typing import Set

from azure.ai.projects import AIProjectClient
from azure.ai.projects.models import ToolSet, FunctionTool
from azure.identity import DefaultAzureCredential

# Example user function
from user_functions import fetch_current_datetime

# Import AzureLogicAppTool and the function factory from user_logic_apps
from user_logic_apps import AzureLogicAppTool, create_send_email_function

# [START register_logic_app]

# Create the project client
project_client = AIProjectClient.from_connection_string(
    credential=DefaultAzureCredential(),
    conn_str=os.environ["PROJECT_CONNECTION_STRING"],
)

# Extract subscription and resource group from the project scope
subscription_id = project_client.scope["subscription_id"]
resource_group = project_client.scope["resource_group_name"]

# Logic App details
logic_app_name = "<LOGIC_APP_NAME>"
trigger_name = "<TRIGGER_NAME>"

# Create and initialize AzureLogicAppTool utility
logic_app_tool = AzureLogicAppTool(subscription_id, resource_group)
logic_app_tool.register_logic_app(logic_app_name, trigger_name)
print(f"Registered logic app '{logic_app_name}' with trigger '{trigger_name}'.")

# Create the specialized "send_email_via_logic_app" function for your agent tools
send_email_func = create_send_email_function(logic_app_tool, logic_app_name)

# Prepare the function tools for the agent
functions_to_use: Set = {
    fetch_current_datetime,
    send_email_func,  # This references the AzureLogicAppTool instance via closure
}
# [END register_logic_app]

with project_client:
    # Create an agent
    functions = FunctionTool(functions=functions_to_use)
    toolset = ToolSet()
    toolset.add(functions)

    agent = project_client.agents.create_agent(
        model=os.environ["MODEL_DEPLOYMENT_NAME"],
        name="SendEmailAgent",
        instructions="You are a specialized agent for sending emails.",
        toolset=toolset,
    )
    print(f"Created agent, ID: {agent.id}")

    # Create a thread for communication
    thread = project_client.agents.create_thread()
    print(f"Created thread, ID: {thread.id}")

    # Create a message in the thread
    message = project_client.agents.create_message(
        thread_id=thread.id,
        role="user",
        content="Hello, please send an email to <RECIPIENT_EMAIL> with the date and time in '%Y-%m-%d %H:%M:%S' format.",
    )
    print(f"Created message, ID: {message.id}")

    # Create and process an agent run in the thread
    run = project_client.agents.create_and_process_run(thread_id=thread.id, agent_id=agent.id)
    print(f"Run finished with status: {run.status}")

    if run.status == "failed":
        print(f"Run failed: {run.last_error}")

    # Delete the agent when done
    project_client.agents.delete_agent(agent.id)
    print("Deleted agent")

    # Fetch and log all messages
    messages = project_client.agents.list_messages(thread_id=thread.id)
    print(f"Messages: {messages}")



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=wrong-import-position

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._patch import *  # pylint: disable=unused-wildcard-import


from ._models import (  # type: ignore
    AISearchIndexResource,
    AOAIModelConfig,
    Agent,
    AgentDeletionStatus,
    AgentThread,
    AgentThreadCreationOptions,
    AgentsApiResponseFormat,
    AgentsNamedToolChoice,
    ApplicationInsightsConfiguration,
    AzureAISearchResource,
    AzureAISearchToolDefinition,
    AzureFunctionBinding,
    AzureFunctionDefinition,
    AzureFunctionStorageQueue,
    AzureFunctionToolDefinition,
    BingCustomSearchToolDefinition,
    BingGroundingToolDefinition,
    CodeInterpreterToolDefinition,
    CodeInterpreterToolResource,
    ConnectedAgentDetails,
    ConnectedAgentToolDefinition,
    CronTrigger,
    Dataset,
    Evaluation,
    EvaluationSchedule,
    EvaluationTarget,
    EvaluatorConfiguration,
    FileDeletionStatus,
    FileListResponse,
    FileSearchRankingOptions,
    FileSearchToolCallContent,
    FileSearchToolDefinition,
    FileSearchToolDefinitionDetails,
    FileSearchToolResource,
    FunctionDefinition,
    FunctionName,
    FunctionToolDefinition,
    IncompleteRunDetails,
    InputData,
    MAASModelConfig,
    MessageAttachment,
    MessageContent,
    MessageDelta,
    MessageDeltaChunk,
    MessageDeltaContent,
    MessageDeltaImageFileContent,
    MessageDeltaImageFileContentObject,
    MessageDeltaTextAnnotation,
    MessageDeltaTextContent,
    MessageDeltaTextContentObject,
    MessageDeltaTextFileCitationAnnotation,
    MessageDeltaTextFileCitationAnnotationObject,
    MessageDeltaTextFilePathAnnotation,
    MessageDeltaTextFilePathAnnotationObject,
    MessageDeltaTextUrlCitationAnnotation,
    MessageDeltaTextUrlCitationDetails,
    MessageImageFileContent,
    MessageImageFileDetails,
    MessageImageFileParam,
    MessageImageUrlParam,
    MessageIncompleteDetails,
    MessageInputContentBlock,
    MessageInputImageFileBlock,
    MessageInputImageUrlBlock,
    MessageInputTextBlock,
    MessageTextAnnotation,
    MessageTextContent,
    MessageTextDetails,
    MessageTextFileCitationAnnotation,
    MessageTextFileCitationDetails,
    MessageTextFilePathAnnotation,
    MessageTextFilePathDetails,
    MessageTextUrlCitationAnnotation,
    MessageTextUrlCitationDetails,
    MicrosoftFabricToolDefinition,
    OpenAIFile,
    OpenAIPageableListOfAgent,
    OpenAIPageableListOfAgentThread,
    OpenAIPageableListOfRunStep,
    OpenAIPageableListOfThreadMessage,
    OpenAIPageableListOfThreadRun,
    OpenAIPageableListOfVectorStore,
    OpenAIPageableListOfVectorStoreFile,
    OpenApiAnonymousAuthDetails,
    OpenApiAuthDetails,
    OpenApiConnectionAuthDetails,
    OpenApiConnectionSecurityScheme,
    OpenApiFunctionDefinition,
    OpenApiManagedAuthDetails,
    OpenApiManagedSecurityScheme,
    OpenApiToolDefinition,
    RecurrenceSchedule,
    RecurrenceTrigger,
    RequiredAction,
    RequiredFunctionToolCall,
    RequiredFunctionToolCallDetails,
    RequiredToolCall,
    ResponseFormatJsonSchema,
    ResponseFormatJsonSchemaType,
    RunCompletionUsage,
    RunError,
    RunStep,
    RunStepAzureAISearchToolCall,
    RunStepBingGroundingToolCall,
    RunStepCodeInterpreterImageOutput,
    RunStepCodeInterpreterImageReference,
    RunStepCodeInterpreterLogOutput,
    RunStepCodeInterpreterToolCall,
    RunStepCodeInterpreterToolCallDetails,
    RunStepCodeInterpreterToolCallOutput,
    RunStepCompletionUsage,
    RunStepCustomSearchToolCall,
    RunStepDelta,
    RunStepDeltaChunk,
    RunStepDeltaCodeInterpreterDetailItemObject,
    RunStepDeltaCodeInterpreterImageOutput,
    RunStepDeltaCodeInterpreterImageOutputObject,
    RunStepDeltaCodeInterpreterLogOutput,
    RunStepDeltaCodeInterpreterOutput,
    RunStepDeltaCodeInterpreterToolCall,
    RunStepDeltaDetail,
    RunStepDeltaFileSearchToolCall,
    RunStepDeltaFunction,
    RunStepDeltaFunctionToolCall,
    RunStepDeltaMessageCreation,
    RunStepDeltaMessageCreationObject,
    RunStepDeltaToolCall,
    RunStepDeltaToolCallObject,
    RunStepDetails,
    RunStepError,
    RunStepFileSearchToolCall,
    RunStepFileSearchToolCallResult,
    RunStepFileSearchToolCallResults,
    RunStepFunctionToolCall,
    RunStepFunctionToolCallDetails,
    RunStepMessageCreationDetails,
    RunStepMessageCreationReference,
    RunStepMicrosoftFabricToolCall,
    RunStepOpenAPIToolCall,
    RunStepSharepointToolCall,
    RunStepToolCall,
    RunStepToolCallDetails,
    SearchConfiguration,
    SearchConfigurationList,
    SharepointToolDefinition,
    SubmitToolOutputsAction,
    SubmitToolOutputsDetails,
    SystemData,
    TargetModelConfig,
    ThreadDeletionStatus,
    ThreadMessage,
    ThreadMessageOptions,
    ThreadRun,
    ToolConnection,
    ToolConnectionList,
    ToolDefinition,
    ToolOutput,
    ToolResources,
    Trigger,
    TruncationObject,
    UpdateCodeInterpreterToolResourceOptions,
    UpdateFileSearchToolResourceOptions,
    UpdateToolResourcesOptions,
    VectorStore,
    VectorStoreAutoChunkingStrategyRequest,
    VectorStoreAutoChunkingStrategyResponse,
    VectorStoreChunkingStrategyRequest,
    VectorStoreChunkingStrategyResponse,
    VectorStoreConfiguration,
    VectorStoreConfigurations,
    VectorStoreDataSource,
    VectorStoreDeletionStatus,
    VectorStoreExpirationPolicy,
    VectorStoreFile,
    VectorStoreFileBatch,
    VectorStoreFileCount,
    VectorStoreFileDeletionStatus,
    VectorStoreFileError,
    VectorStoreStaticChunkingStrategyOptions,
    VectorStoreStaticChunkingStrategyRequest,
    VectorStoreStaticChunkingStrategyResponse,
)

from ._enums import (  # type: ignore
    AgentStreamEvent,
    AgentsApiResponseFormatMode,
    AgentsApiToolChoiceOptionMode,
    AgentsNamedToolChoiceType,
    AuthenticationType,
    AzureAISearchQueryType,
    ConnectionType,
    DoneEvent,
    ErrorEvent,
    FilePurpose,
    FileState,
    Frequency,
    ImageDetailLevel,
    IncompleteDetailsReason,
    ListSortOrder,
    MessageBlockType,
    MessageIncompleteDetailsReason,
    MessageRole,
    MessageStatus,
    MessageStreamEvent,
    OpenApiAuthType,
    ResponseFormat,
    RunAdditionalFieldList,
    RunStatus,
    RunStepErrorCode,
    RunStepStatus,
    RunStepStreamEvent,
    RunStepType,
    RunStreamEvent,
    ThreadStreamEvent,
    TruncationStrategy,
    VectorStoreChunkingStrategyRequestType,
    VectorStoreChunkingStrategyResponseType,
    VectorStoreDataSourceAssetType,
    VectorStoreExpirationPolicyAnchor,
    VectorStoreFileBatchStatus,
    VectorStoreFileErrorCode,
    VectorStoreFileStatus,
    VectorStoreFileStatusFilter,
    VectorStoreStatus,
    WeekDays,
)
from ._patch import __all__ as _patch_all
from ._patch import *
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "AISearchIndexResource",
    "AOAIModelConfig",
    "Agent",
    "AgentDeletionStatus",
    "AgentThread",
    "AgentThreadCreationOptions",
    "AgentsApiResponseFormat",
    "AgentsNamedToolChoice",
    "ApplicationInsightsConfiguration",
    "AzureAISearchResource",
    "AzureAISearchToolDefinition",
    "AzureFunctionBinding",
    "AzureFunctionDefinition",
    "AzureFunctionStorageQueue",
    "AzureFunctionToolDefinition",
    "BingCustomSearchToolDefinition",
    "BingGroundingToolDefinition",
    "CodeInterpreterToolDefinition",
    "CodeInterpreterToolResource",
    "ConnectedAgentDetails",
    "ConnectedAgentToolDefinition",
    "CronTrigger",
    "Dataset",
    "Evaluation",
    "EvaluationSchedule",
    "EvaluationTarget",
    "EvaluatorConfiguration",
    "FileDeletionStatus",
    "FileListResponse",
    "FileSearchRankingOptions",
    "FileSearchToolCallContent",
    "FileSearchToolDefinition",
    "FileSearchToolDefinitionDetails",
    "FileSearchToolResource",
    "FunctionDefinition",
    "FunctionName",
    "FunctionToolDefinition",
    "IncompleteRunDetails",
    "InputData",
    "MAASModelConfig",
    "MessageAttachment",
    "MessageContent",
    "MessageDelta",
    "MessageDeltaChunk",
    "MessageDeltaContent",
    "MessageDeltaImageFileContent",
    "MessageDeltaImageFileContentObject",
    "MessageDeltaTextAnnotation",
    "MessageDeltaTextContent",
    "MessageDeltaTextContentObject",
    "MessageDeltaTextFileCitationAnnotation",
    "MessageDeltaTextFileCitationAnnotationObject",
    "MessageDeltaTextFilePathAnnotation",
    "MessageDeltaTextFilePathAnnotationObject",
    "MessageDeltaTextUrlCitationAnnotation",
    "MessageDeltaTextUrlCitationDetails",
    "MessageImageFileContent",
    "MessageImageFileDetails",
    "MessageImageFileParam",
    "MessageImageUrlParam",
    "MessageIncompleteDetails",
    "MessageInputContentBlock",
    "MessageInputImageFileBlock",
    "MessageInputImageUrlBlock",
    "MessageInputTextBlock",
    "MessageTextAnnotation",
    "MessageTextContent",
    "MessageTextDetails",
    "MessageTextFileCitationAnnotation",
    "MessageTextFileCitationDetails",
    "MessageTextFilePathAnnotation",
    "MessageTextFilePathDetails",
    "MessageTextUrlCitationAnnotation",
    "MessageTextUrlCitationDetails",
    "MicrosoftFabricToolDefinition",
    "OpenAIFile",
    "OpenAIPageableListOfAgent",
    "OpenAIPageableListOfAgentThread",
    "OpenAIPageableListOfRunStep",
    "OpenAIPageableListOfThreadMessage",
    "OpenAIPageableListOfThreadRun",
    "OpenAIPageableListOfVectorStore",
    "OpenAIPageableListOfVectorStoreFile",
    "OpenApiAnonymousAuthDetails",
    "OpenApiAuthDetails",
    "OpenApiConnectionAuthDetails",
    "OpenApiConnectionSecurityScheme",
    "OpenApiFunctionDefinition",
    "OpenApiManagedAuthDetails",
    "OpenApiManagedSecurityScheme",
    "OpenApiToolDefinition",
    "RecurrenceSchedule",
    "RecurrenceTrigger",
    "RequiredAction",
    "RequiredFunctionToolCall",
    "RequiredFunctionToolCallDetails",
    "RequiredToolCall",
    "ResponseFormatJsonSchema",
    "ResponseFormatJsonSchemaType",
    "RunCompletionUsage",
    "RunError",
    "RunStep",
    "RunStepAzureAISearchToolCall",
    "RunStepBingGroundingToolCall",
    "RunStepCodeInterpreterImageOutput",
    "RunStepCodeInterpreterImageReference",
    "RunStepCodeInterpreterLogOutput",
    "RunStepCodeInterpreterToolCall",
    "RunStepCodeInterpreterToolCallDetails",
    "RunStepCodeInterpreterToolCallOutput",
    "RunStepCompletionUsage",
    "RunStepCustomSearchToolCall",
    "RunStepDelta",
    "RunStepDeltaChunk",
    "RunStepDeltaCodeInterpreterDetailItemObject",
    "RunStepDeltaCodeInterpreterImageOutput",
    "RunStepDeltaCodeInterpreterImageOutputObject",
    "RunStepDeltaCodeInterpreterLogOutput",
    "RunStepDeltaCodeInterpreterOutput",
    "RunStepDeltaCodeInterpreterToolCall",
    "RunStepDeltaDetail",
    "RunStepDeltaFileSearchToolCall",
    "RunStepDeltaFunction",
    "RunStepDeltaFunctionToolCall",
    "RunStepDeltaMessageCreation",
    "RunStepDeltaMessageCreationObject",
    "RunStepDeltaToolCall",
    "RunStepDeltaToolCallObject",
    "RunStepDetails",
    "RunStepError",
    "RunStepFileSearchToolCall",
    "RunStepFileSearchToolCallResult",
    "RunStepFileSearchToolCallResults",
    "RunStepFunctionToolCall",
    "RunStepFunctionToolCallDetails",
    "RunStepMessageCreationDetails",
    "RunStepMessageCreationReference",
    "RunStepMicrosoftFabricToolCall",
    "RunStepOpenAPIToolCall",
    "RunStepSharepointToolCall",
    "RunStepToolCall",
    "RunStepToolCallDetails",
    "SearchConfiguration",
    "SearchConfigurationList",
    "SharepointToolDefinition",
    "SubmitToolOutputsAction",
    "SubmitToolOutputsDetails",
    "SystemData",
    "TargetModelConfig",
    "ThreadDeletionStatus",
    "ThreadMessage",
    "ThreadMessageOptions",
    "ThreadRun",
    "ToolConnection",
    "ToolConnectionList",
    "ToolDefinition",
    "ToolOutput",
    "ToolResources",
    "Trigger",
    "TruncationObject",
    "UpdateCodeInterpreterToolResourceOptions",
    "UpdateFileSearchToolResourceOptions",
    "UpdateToolResourcesOptions",
    "VectorStore",
    "VectorStoreAutoChunkingStrategyRequest",
    "VectorStoreAutoChunkingStrategyResponse",
    "VectorStoreChunkingStrategyRequest",
    "VectorStoreChunkingStrategyResponse",
    "VectorStoreConfiguration",
    "VectorStoreConfigurations",
    "VectorStoreDataSource",
    "VectorStoreDeletionStatus",
    "VectorStoreExpirationPolicy",
    "VectorStoreFile",
    "VectorStoreFileBatch",
    "VectorStoreFileCount",
    "VectorStoreFileDeletionStatus",
    "VectorStoreFileError",
    "VectorStoreStaticChunkingStrategyOptions",
    "VectorStoreStaticChunkingStrategyRequest",
    "VectorStoreStaticChunkingStrategyResponse",
    "AgentStreamEvent",
    "AgentsApiResponseFormatMode",
    "AgentsApiToolChoiceOptionMode",
    "AgentsNamedToolChoiceType",
    "AuthenticationType",
    "AzureAISearchQueryType",
    "ConnectionType",
    "DoneEvent",
    "ErrorEvent",
    "FilePurpose",
    "FileState",
    "Frequency",
    "ImageDetailLevel",
    "IncompleteDetailsReason",
    "ListSortOrder",
    "MessageBlockType",
    "MessageIncompleteDetailsReason",
    "MessageRole",
    "MessageStatus",
    "MessageStreamEvent",
    "OpenApiAuthType",
    "ResponseFormat",
    "RunAdditionalFieldList",
    "RunStatus",
    "RunStepErrorCode",
    "RunStepStatus",
    "RunStepStreamEvent",
    "RunStepType",
    "RunStreamEvent",
    "ThreadStreamEvent",
    "TruncationStrategy",
    "VectorStoreChunkingStrategyRequestType",
    "VectorStoreChunkingStrategyResponseType",
    "VectorStoreDataSourceAssetType",
    "VectorStoreExpirationPolicyAnchor",
    "VectorStoreFileBatchStatus",
    "VectorStoreFileErrorCode",
    "VectorStoreFileStatus",
    "VectorStoreFileStatusFilter",
    "VectorStoreStatus",
    "WeekDays",
]
__all__.extend([p for p in _patch_all if p not in __all__])  # pyright: ignore
_patch_sdk()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import functools
import asyncio  # pylint: disable=do-not-import-asyncio
from typing import Any, Callable, Optional, Dict

try:
    # pylint: disable = no-name-in-module
    from opentelemetry import trace as opentelemetry_trace

    tracer = opentelemetry_trace.get_tracer(__name__)
    _tracing_library_available = True
except ModuleNotFoundError:
    _tracing_library_available = False

if _tracing_library_available:

    def trace_function(span_name: Optional[str] = None):
        """
        A decorator for tracing function calls using OpenTelemetry.

        This decorator handles various data types for function parameters and return values,
        and records them as attributes in the trace span. The supported data types include:
        - Basic data types: str, int, float, bool
        - Collections: list, dict, tuple, set

        Special handling for collections:
        - If a collection (list, dict, tuple, set) contains nested collections, the entire collection
        is converted to a string before being recorded as an attribute.
        - Sets and dictionaries are always converted to strings to ensure compatibility with span attributes.

        Object types are omitted, and the corresponding parameter is not traced.

        :param span_name: The name of the span. If not provided, the function name is used.
        :type span_name: Optional[str]
        :return: The decorated function with tracing enabled.
        :rtype: Callable
        """

        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                Wrapper function for asynchronous functions.

                :param args: Positional arguments passed to the function.
                :type args: Tuple[Any]
                :return: The result of the decorated asynchronous function.
                :rtype: Any
                """
                name = span_name if span_name else func.__name__
                with tracer.start_as_current_span(name) as span:
                    try:
                        # Sanitize parameters and set them as attributes
                        sanitized_params = sanitize_parameters(func, *args, **kwargs)
                        span.set_attributes(sanitized_params)
                        result = await func(*args, **kwargs)
                        sanitized_result = sanitize_for_attributes(result)
                        if sanitized_result is not None:
                            if isinstance(sanitized_result, (list, dict, tuple, set)):
                                if any(isinstance(i, (list, dict, tuple, set)) for i in sanitized_result):
                                    sanitized_result = str(sanitized_result)
                            span.set_attribute("code.function.return.value", sanitized_result)  # type: ignore
                        return result
                    except Exception as e:
                        span.record_exception(e)
                        span.set_attribute("error.type", e.__class__.__qualname__)  # type: ignore
                        raise

            @functools.wraps(func)
            def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                Wrapper function for synchronous functions.

                :param args: Positional arguments passed to the function.
                :type args: Tuple[Any]
                :return: The result of the decorated synchronous function.
                :rtype: Any
                """
                name = span_name if span_name else func.__name__
                with tracer.start_as_current_span(name) as span:
                    try:
                        # Sanitize parameters and set them as attributes
                        sanitized_params = sanitize_parameters(func, *args, **kwargs)
                        span.set_attributes(sanitized_params)
                        result = func(*args, **kwargs)
                        sanitized_result = sanitize_for_attributes(result)
                        if sanitized_result is not None:
                            if isinstance(sanitized_result, (list, dict, tuple, set)):
                                if any(isinstance(i, (list, dict, tuple, set)) for i in sanitized_result):
                                    sanitized_result = str(sanitized_result)
                            span.set_attribute("code.function.return.value", sanitized_result)  # type: ignore
                        return result
                    except Exception as e:
                        span.record_exception(e)
                        span.set_attribute("error.type", e.__class__.__qualname__)  # type: ignore
                        raise

            # Determine if the function is async
            if asyncio.iscoroutinefunction(func):
                return async_wrapper
            return sync_wrapper

        return decorator

else:
    # Define a no-op decorator if OpenTelemetry is not available
    def trace_function(span_name: Optional[str] = None):  # pylint: disable=unused-argument
        """
        A no-op decorator for tracing function calls when OpenTelemetry is not available.

        :param span_name: Not used in this version.
        :type span_name: Optional[str]
        :return: The original function.
        :rtype: Callable
        """

        def decorator(func: Callable) -> Callable:
            return func

        return decorator


def sanitize_parameters(func, *args, **kwargs) -> Dict[str, Any]:
    """
    Sanitize function parameters to include only built-in data types.

    :param func: The function being decorated.
    :type func: Callable
    :param args: Positional arguments passed to the function.
    :type args: Tuple[Any]
    :return: A dictionary of sanitized parameters.
    :rtype: Dict[str, Any]
    """
    import inspect

    params = inspect.signature(func).parameters
    sanitized_params = {}

    for i, (name, param) in enumerate(params.items()):
        if param.default == inspect.Parameter.empty and i < len(args):
            value = args[i]
        else:
            value = kwargs.get(name, param.default)

        sanitized_value = sanitize_for_attributes(value)
        # Check if the collection has nested collections
        if isinstance(sanitized_value, (list, dict, tuple, set)):
            if any(isinstance(i, (list, dict, tuple, set)) for i in sanitized_value):
                sanitized_value = str(sanitized_value)
        if sanitized_value is not None:
            sanitized_params["code.function.parameter." + name] = sanitized_value

    return sanitized_params


# pylint: disable=R0911
def sanitize_for_attributes(value: Any, is_recursive: bool = False) -> Any:
    """
    Sanitize a value to be used as an attribute.

    :param value: The value to sanitize.
    :type value: Any
    :param is_recursive: Indicates if the function is being called recursively. Default is False.
    :type is_recursive: bool
    :return: The sanitized value or None if the value is not a supported type.
    :rtype: Any
    """
    if isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, list):
        return [
            sanitize_for_attributes(item, True)
            for item in value
            if isinstance(item, (str, int, float, bool, list, dict, tuple, set))
        ]
    if isinstance(value, dict):
        retval = {
            k: sanitize_for_attributes(v, True)
            for k, v in value.items()
            if isinstance(v, (str, int, float, bool, list, dict, tuple, set))
        }
        # dict to compatible with span attribute, so return it as a string
        if is_recursive:
            return retval
        return str(retval)
    if isinstance(value, tuple):
        return tuple(
            sanitize_for_attributes(item, True)
            for item in value
            if isinstance(item, (str, int, float, bool, list, dict, tuple, set))
        )
    if isinstance(value, set):
        retval_set = {
            sanitize_for_attributes(item, True)
            for item in value
            if isinstance(item, (str, int, float, bool, list, dict, tuple, set))
        }
        if is_recursive:
            return retval_set
        return str(retval_set)
    return None



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use agent operations with file searching from
    the Azure Agents service using a synchronous client.

USAGE:
    python sample_agents_file_search.py

    Before running the sample:

    pip install azure-ai-projects azure-identity

    Set these environment variables with your own values:
    1) PROJECT_CONNECTION_STRING - The project connection string, as found in the overview page of your
       Azure AI Foundry project.
    2) MODEL_DEPLOYMENT_NAME - The deployment name of the AI model, as found under the "Name" column in 
       the "Models + endpoints" tab in your Azure AI Foundry project.
"""

import os
from azure.ai.projects import AIProjectClient
from azure.ai.projects.models import (
    FileSearchTool,
)
from azure.identity import DefaultAzureCredential

project_client = AIProjectClient.from_connection_string(
    credential=DefaultAzureCredential(), conn_str=os.environ["PROJECT_CONNECTION_STRING"]
)

with project_client:

    # Upload file and create vector store
    # [START upload_file_create_vector_store_and_agent_with_file_search_tool]
    file = project_client.agents.upload_file_and_poll(file_path="product_info_1.md", purpose="assistants")
    print(f"Uploaded file, file ID: {file.id}")

    vector_store = project_client.agents.create_vector_store_and_poll(file_ids=[file.id], name="my_vectorstore")
    print(f"Created vector store, vector store ID: {vector_store.id}")

    # Create file search tool with resources followed by creating agent
    file_search = FileSearchTool(vector_store_ids=[vector_store.id])

    agent = project_client.agents.create_agent(
        model=os.environ["MODEL_DEPLOYMENT_NAME"],
        name="my-assistant",
        instructions="Hello, you are helpful assistant and can search information from uploaded files",
        tools=file_search.definitions,
        tool_resources=file_search.resources,
    )
    # [END upload_file_create_vector_store_and_agent_with_file_search_tool]

    print(f"Created agent, ID: {agent.id}")

    # Create thread for communication
    thread = project_client.agents.create_thread()
    print(f"Created thread, ID: {thread.id}")

    # Create message to thread
    message = project_client.agents.create_message(
        thread_id=thread.id, role="user", content="Hello, what Contoso products do you know?"
    )
    print(f"Created message, ID: {message.id}")

    # Create and process assistant run in thread with tools
    run = project_client.agents.create_and_process_run(thread_id=thread.id, agent_id=agent.id)
    print(f"Run finished with status: {run.status}")

    if run.status == "failed":
        # Check if you got "Rate limit is exceeded.", then you want to get more quota
        print(f"Run failed: {run.last_error}")

    # [START teardown]
    # Delete the file when done
    project_client.agents.delete_vector_store(vector_store.id)
    print("Deleted vector store")

    project_client.agents.delete_file(file_id=file.id)
    print("Deleted file")

    # Delete the agent when done
    project_client.agents.delete_agent(agent.id)
    print("Deleted agent")
    # [END teardown]

    # Fetch and log all messages
    messages = project_client.agents.list_messages(thread_id=thread.id)

    # Print messages from the thread
    for text_message in messages.text_messages:
        print(text_message)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import numbers
import json
from typing import List
from opentelemetry.sdk.trace import Span


class GenAiTraceVerifier:

    def check_span_attributes(self, span, attributes):
        # Convert the list of tuples to a dictionary for easier lookup
        attribute_dict = dict(attributes)

        for attribute_name in span.attributes.keys():
            # Check if the attribute name exists in the input attributes
            if attribute_name not in attribute_dict:
                print("Attribute name " + attribute_name + " not in attribute dictionary")
                return False

            attribute_value = attribute_dict[attribute_name]
            if isinstance(attribute_value, list):
                # Check if the attribute value in the span matches the provided list
                if span.attributes[attribute_name] != attribute_value:
                    print(
                        "Attribute value list "
                        + str(span.attributes[attribute_name])
                        + " does not match with "
                        + str(attribute_value)
                    )
                    return False
            elif isinstance(attribute_value, tuple):
                # Check if the attribute value in the span matches the provided list
                if span.attributes[attribute_name] != attribute_value:
                    print(
                        "Attribute value tuple "
                        + str(span.attributes[attribute_name])
                        + " does not match with "
                        + str(attribute_value)
                    )
                    return False
            else:
                # Check if the attribute value matches the provided value
                if attribute_value == "+":
                    if not isinstance(span.attributes[attribute_name], numbers.Number):
                        print("Attribute value " + str(span.attributes[attribute_name]) + " is not a number")
                        return False
                    if span.attributes[attribute_name] < 0:
                        print("Attribute value " + str(span.attributes[attribute_name]) + " is negative")
                        return False
                elif attribute_value != "" and span.attributes[attribute_name] != attribute_value:
                    print(
                        "Attribute value "
                        + str(span.attributes[attribute_name])
                        + " does not match with "
                        + str(attribute_value)
                    )
                    return False
                # Check if the attribute value in the span is not empty when the provided value is ""
                elif attribute_value == "" and not span.attributes[attribute_name]:
                    print("Expected non-empty attribute value")
                    return False

        return True

    def check_decorator_span_attributes(self, span: Span, attributes: List[tuple]) -> bool:
        # Convert the list of tuples to a dictionary for easier lookup
        attribute_dict = dict(attributes)

        # Ensure all required attributes are present in the span
        for attribute_name in attribute_dict.keys():
            if attribute_name not in span.attributes:
                print("Required attribute name " + attribute_name + " not found in span attributes")
                return False

        for attribute_name in span.attributes.keys():
            # Check if the attribute name exists in the input attributes
            if attribute_name not in attribute_dict:
                print("Attribute name " + attribute_name + " not in attribute dictionary")
                return False

            attribute_value = attribute_dict[attribute_name]
            span_value = span.attributes[attribute_name]

            if isinstance(attribute_value, (list, tuple)):
                # Convert both to lists for comparison
                if list(span_value) != list(attribute_value):
                    print(
                        "Attribute value list/tuple " + str(span_value) + " does not match with " + str(attribute_value)
                    )
                    return False
            elif isinstance(attribute_value, dict):
                # Check if both are dictionaries and compare them
                if not isinstance(span_value, dict) or span_value != attribute_value:
                    print("Attribute value dict " + str(span_value) + " does not match with " + str(attribute_value))
                    return False
            else:
                # Check if the attribute value matches the provided value
                if attribute_value == "+":
                    if not isinstance(span_value, numbers.Number):
                        print("Attribute value " + str(span_value) + " is not a number")
                        return False
                    if span_value < 0:
                        print("Attribute value " + str(span_value) + " is negative")
                        return False
                elif attribute_value != "" and span_value != attribute_value:
                    print("Attribute value " + str(span_value) + " does not match with " + str(attribute_value))
                    return False
                # Check if the attribute value in the span is not empty when the provided value is ""
                elif attribute_value == "" and not span_value:
                    print("Expected non-empty attribute value")
                    return False

        return True

    def is_valid_json(self, my_string):
        try:
            json_object = json.loads(my_string)
        except ValueError as e1:
            return False
        except TypeError as e2:
            return False
        return True

    def check_json_string(self, expected_json, actual_json):
        if self.is_valid_json(expected_json) and self.is_valid_json(actual_json):
            return self.check_event_attributes(json.loads(expected_json), json.loads(actual_json))
        else:
            return False

    def check_event_attributes(self, expected_dict, actual_dict):
        if set(expected_dict.keys()) != set(actual_dict.keys()):
            if isinstance(expected_dict, dict):
                expected_val = json.dumps(expected_dict)
            else:
                expected_val = expected_dict
            if isinstance(actual_dict, dict):
                actual_val = json.dumps(actual_dict)
            else:
                actual_val = actual_dict
            print("check_event_attributes: keys do not match: " + expected_val + "!=" + actual_val)
            return False
        for key, expected_val in expected_dict.items():
            if key not in actual_dict:
                print("check_event_attributes: key not found")
                return False
            actual_val = actual_dict[key]

            if self.is_valid_json(expected_val):
                if not self.is_valid_json(actual_val):
                    print("check_event_attributes: actual_val is not valid json")
                    return False
                if not self.check_json_string(expected_val, actual_val):
                    print("check_event_attributes: check_json_string failed")
                    return False
            elif isinstance(expected_val, dict):
                if not isinstance(actual_val, dict):
                    print("check_event_attributes: actual_val is not dict")
                    return False
                if not self.check_event_attributes(expected_val, actual_val):
                    print("check_event_attributes: check_event_attributes failed")
                    return False
            elif isinstance(expected_val, list):
                if not isinstance(actual_val, list):
                    print("check_event_attributes: actual_val is not list")
                    return False
                if len(expected_val) != len(actual_val):
                    print("check_event_attributes: list lengths do not match")
                    return False
                for expected_list, actual_list in zip(expected_val, actual_val):
                    if not self.check_event_attributes(expected_list, actual_list):
                        print("check_event_attributes: check_event_attributes for list failed")
                        return False
            elif isinstance(expected_val, str) and expected_val == "*":
                if actual_val == "":
                    print("check_event_attributes: actual_val is empty")
                    return False
            elif isinstance(expected_val, str) and expected_val == "+":
                if not isinstance(actual_val, numbers.Number):
                    return False
                if actual_val < 0:
                    return False
            elif expected_val != actual_val:
                if isinstance(expected_val, dict):
                    expected_val = json.dumps(expected_val)
                if isinstance(actual_val, dict):
                    actual_val = json.dumps(actual_val)
                print("check_event_attributes: values do not match: " + expected_val + "!=" + actual_val)
                return False
        return True

    def check_span_events(self, span, expected_events):
        print("Checking span: " + span.name)
        span_events = list(span.events)  # Create a list of events from the span

        for expected_event in expected_events:
            for actual_event in span_events:
                if expected_event["name"] == actual_event.name:
                    if not self.check_event_attributes(expected_event["attributes"], actual_event.attributes._dict):
                        print("Event attributes do not match")
                        return False
                    span_events.remove(actual_event)  # Remove the matched event from the span_events
                    break
            else:
                print("check_span_events: event not found")
                return False  # If no match found for an expected event

        if len(span_events) > 0:  # If there are any additional events in the span_events
            print("check_span_events: unexpected event found")
            return False

        return True



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use agent operations with an event handler and toolset from
    the Azure Agents service using a asynchronous client.

USAGE:
    python sample_agents_stream_eventhandler_with_toolset_async.py

    Before running the sample:

    pip install azure-ai-projects azure-identity aiohttp

    Set this environment variables with your own values:
    PROJECT_CONNECTION_STRING - the Azure AI Project connection string, as found in your AI Foundry project.
"""
import asyncio
from typing import Any

from azure.ai.projects.aio import AIProjectClient
from azure.ai.projects.models import MessageDeltaChunk, RunStep, ThreadMessage, ThreadRun
from azure.ai.projects.models import AsyncAgentEventHandler, AsyncFunctionTool, AsyncToolSet
from azure.identity.aio import DefaultAzureCredential

import os

from user_async_functions import user_async_functions


class MyEventHandler(AsyncAgentEventHandler):

    async def on_message_delta(self, delta: "MessageDeltaChunk") -> None:
        print(f"Text delta received: {delta.text}")

    async def on_thread_message(self, message: "ThreadMessage") -> None:
        print(f"ThreadMessage created. ID: {message.id}, Status: {message.status}")

    async def on_thread_run(self, run: "ThreadRun") -> None:
        print(f"ThreadRun status: {run.status}")

        if run.status == "failed":
            print(f"Run failed. Error: {run.last_error}")

    async def on_run_step(self, step: "RunStep") -> None:
        print(f"RunStep type: {step.type}, Status: {step.status}")

    async def on_error(self, data: str) -> None:
        print(f"An error occurred. Data: {data}")

    async def on_done(self) -> None:
        print("Stream completed.")

    async def on_unhandled_event(self, event_type: str, event_data: Any) -> None:
        print(f"Unhandled Event Type: {event_type}, Data: {event_data}")


async def main() -> None:
    async with DefaultAzureCredential() as creds:
        async with AIProjectClient.from_connection_string(
            credential=creds, conn_str=os.environ["PROJECT_CONNECTION_STRING"]
        ) as project_client:

            # Initialize toolset with user functions
            functions = AsyncFunctionTool(user_async_functions)
            toolset = AsyncToolSet()
            toolset.add(functions)

            project_client.agents.enable_auto_function_calls(functions=user_async_functions)
            agent = await project_client.agents.create_agent(
                model=os.environ["MODEL_DEPLOYMENT_NAME"],
                name="my-assistant",
                instructions="You are helpful assistant",
                toolset=toolset,
            )
            print(f"Created agent, agent ID: {agent.id}")

            thread = await project_client.agents.create_thread()
            print(f"Created thread, thread ID {thread.id}")

            message = await project_client.agents.create_message(
                thread_id=thread.id,
                role="user",
                content="Hello, send an email with the datetime and weather information in New York? Also let me know the details",
            )
            print(f"Created message, message ID {message.id}")

            async with await project_client.agents.create_stream(
                thread_id=thread.id, agent_id=agent.id, event_handler=MyEventHandler()
            ) as stream:
                await stream.until_done()

            await project_client.agents.delete_agent(agent.id)
            print("Deleted agent")

            messages = await project_client.agents.list_messages(thread_id=thread.id)
            print(f"Messages: {messages}")


if __name__ == "__main__":
    asyncio.run(main())



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to use agent operations to add files to an existing vector store and perform search from
    the Azure Agents service using a asynchronous client.

USAGE:
    python sample_agents_vector_store_batch_file_search_async.py

    Before running the sample:

    pip install azure-ai-projects azure-identity aiohttp

    Set this environment variables with your own values:
    PROJECT_CONNECTION_STRING - the Azure AI Project connection string, as found in your AI Foundry project.
"""

import asyncio
import os
from azure.ai.projects.aio import AIProjectClient
from azure.ai.projects.models import FileSearchTool, FilePurpose
from azure.identity.aio import DefaultAzureCredential


async def main() -> None:
    async with DefaultAzureCredential() as creds:
        async with AIProjectClient.from_connection_string(
            credential=creds, conn_str=os.environ["PROJECT_CONNECTION_STRING"]
        ) as project_client:
            # Upload a file and wait for it to be processed
            file = await project_client.agents.upload_file_and_poll(
                file_path="../product_info_1.md", purpose=FilePurpose.AGENTS
            )
            print(f"Uploaded file, file ID: {file.id}")

            # Create a vector store with no file and wait for it to be processed
            vector_store = await project_client.agents.create_vector_store_and_poll(
                file_ids=[], name="sample_vector_store"
            )
            print(f"Created vector store, vector store ID: {vector_store.id}")

            # Add the file to the vector store or you can supply file ids in the vector store creation
            vector_store_file_batch = await project_client.agents.create_vector_store_file_batch_and_poll(
                vector_store_id=vector_store.id, file_ids=[file.id]
            )
            print(f"Created vector store file batch, vector store file batch ID: {vector_store_file_batch.id}")

            # Create a file search tool
            file_search_tool = FileSearchTool(vector_store_ids=[vector_store.id])

            # Notices that FileSearchTool as tool and tool_resources must be added or the assistant unable to search the file
            agent = await project_client.agents.create_agent(
                model=os.environ["MODEL_DEPLOYMENT_NAME"],
                name="my-assistant",
                instructions="You are helpful assistant",
                tools=file_search_tool.definitions,
                tool_resources=file_search_tool.resources,
            )
            print(f"Created agent, agent ID: {agent.id}")

            thread = await project_client.agents.create_thread()
            print(f"Created thread, thread ID: {thread.id}")

            message = await project_client.agents.create_message(
                thread_id=thread.id, role="user", content="What feature does Smart Eyewear offer?"
            )
            print(f"Created message, message ID: {message.id}")

            run = await project_client.agents.create_and_process_run(thread_id=thread.id, agent_id=agent.id)
            print(f"Created run, run ID: {run.id}")

            file_search_tool.remove_vector_store(vector_store.id)
            print(f"Removed vector store from file search, vector store ID: {vector_store.id}")

            await project_client.agents.update_agent(
                agent_id=agent.id, tools=file_search_tool.definitions, tool_resources=file_search_tool.resources
            )
            print(f"Updated agent, agent ID: {agent.id}")

            thread = await project_client.agents.create_thread()
            print(f"Created thread, thread ID: {thread.id}")

            message = await project_client.agents.create_message(
                thread_id=thread.id, role="user", content="What feature does Smart Eyewear offer?"
            )
            print(f"Created message, message ID: {message.id}")

            run = await project_client.agents.create_and_process_run(thread_id=thread.id, agent_id=agent.id)
            print(f"Created run, run ID: {run.id}")

            await project_client.agents.delete_file(file.id)
            print("Deleted file")

            await project_client.agents.delete_vector_store(vector_store.id)
            print("Deleted vector store")

            await project_client.agents.delete_agent(agent.id)
            print("Deleted agent")

            messages = await project_client.agents.list_messages(thread_id=thread.id)
            print(f"Messages: {messages}")


if __name__ == "__main__":
    asyncio.run(main())



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    Given an AIProjectClient, this sample demonstrates how to get an authenticated 
    async ChatCompletionsClient from the azure.ai.inference package, and then work with a prompt string.
    For more information on the azure.ai.inference package see https://pypi.org/project/azure-ai-inference/.

USAGE:
    python sample_chat_completions_with_azure_ai_inference_client_and_prompt_string.py

    Before running the sample:

    pip install azure-ai-projects azure-identity

    Set these environment variables with your own values:
    * PROJECT_CONNECTION_STRING - The Azure AI Project connection string, as found in your AI Foundry project.
    * MODEL_DEPLOYMENT_NAME - The model deployment name, as found in your AI Foundry project.
"""

import os
from azure.ai.projects import AIProjectClient
from azure.ai.projects.prompts import PromptTemplate
from azure.ai.inference.models import UserMessage
from azure.identity import DefaultAzureCredential

project_connection_string = os.environ["PROJECT_CONNECTION_STRING"]
model_deployment_name = os.environ["MODEL_DEPLOYMENT_NAME"]

with AIProjectClient.from_connection_string(
    credential=DefaultAzureCredential(),
    conn_str=project_connection_string,
) as project_client:

    with project_client.inference.get_chat_completions_client() as client:

        prompt_template_str = """
            system:
            You are an AI assistant in a hotel. You help guests with their requests and provide information about the hotel and its services.

            # context
            {{#rules}}
            {{rule}}
            {{/rules}}

            {{#chat_history}}
            {{role}}:
            {{content}}
            {{/chat_history}}

            user:
            {{input}}
        """
        prompt_template = PromptTemplate.from_string(api="chat", prompt_template=prompt_template_str)

        input = "When I arrived, can I still have breakfast?"
        rules = [
            {"rule": "The check-in time is 3pm"},
            {"rule": "The check-out time is 11am"},
            {"rule": "Breakfast is served from 7am to 10am"},
        ]
        chat_history = [
            {"role": "user", "content": "I'll arrive at 2pm. What's the check-in and check-out time?"},
            {"role": "system", "content": "The check-in time is 3 PM, and the check-out time is 11 AM."},
        ]
        messages = prompt_template.create_messages(input=input, rules=rules, chat_history=chat_history)

        response = client.complete(model=model_deployment_name, messages=messages)

        print(response.choices[0].message.content)



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    This sample demonstrates how to override the base event handler and parse the events and iterate through them
    In your use case, you might not want to write the iteration code similar to sample_agents_stream_iteration_async.py.
    If you have multiple places to call create_stream, you might find the iteration code cumbersome.
    This example shows how to override the base event handler, parse the events, and iterate through them, which can be reused in multiple create_stream calls to help keep the code clean.

USAGE:
    python sample_agents_stream_with_base_override_eventhandler.py

    Before running the sample:

    pip install azure-ai-projects azure-identity

    Set these environment variables with your own values:
    1) PROJECT_CONNECTION_STRING - The project connection string, as found in the overview page of your
       Azure AI Foundry project.
    2) MODEL_DEPLOYMENT_NAME - The deployment name of the AI model, as found under the "Name" column in 
       the "Models + endpoints" tab in your Azure AI Foundry project.
"""
import json
from typing import Generator, Optional

from azure.ai.projects import AIProjectClient
from azure.ai.projects.models import (
    MessageDeltaChunk,
    MessageDeltaTextContent,
)
from azure.ai.projects.models import AgentStreamEvent, BaseAgentEventHandler
from azure.identity import DefaultAzureCredential

import os


# Our goal is to parse the event data in a string and return the chunk in text for each iteration.
# Because we want the iteration to be a string, we define str as the generic type for BaseAsyncAgentEventHandler
# and override the _process_event method to return a string.
# The get_stream_chunks method is defined to return the chunks as strings because the iteration is a string.
class MyEventHandler(BaseAgentEventHandler[Optional[str]]):

    def _process_event(self, event_data_str: str) -> Optional[str]:  # type: ignore[return]
        event_lines = event_data_str.strip().split("\n")
        event_type: Optional[str] = None
        event_data = ""
        for line in event_lines:
            if line.startswith("event:"):
                event_type = line.split(":", 1)[1].strip()
            elif line.startswith("data:"):
                event_data = line.split(":", 1)[1].strip()

        if not event_type:
            raise ValueError("Event type not specified in the event data.")

        if event_type == AgentStreamEvent.THREAD_MESSAGE_DELTA.value:

            event_obj: MessageDeltaChunk = MessageDeltaChunk(**json.loads(event_data))

            for content_part in event_obj.delta.content:
                if isinstance(content_part, MessageDeltaTextContent):
                    if content_part.text is not None:
                        return content_part.text.value
        return None

    def get_stream_chunks(self) -> Generator[str, None, None]:
        for chunk in self:
            if chunk:
                yield chunk


project_client = AIProjectClient.from_connection_string(
    credential=DefaultAzureCredential(), conn_str=os.environ["PROJECT_CONNECTION_STRING"]
)

with project_client:
    agent = project_client.agents.create_agent(
        model=os.environ["MODEL_DEPLOYMENT_NAME"], name="my-assistant", instructions="You are helpful assistant"
    )
    print(f"Created agent, agent ID: {agent.id}")

    thread = project_client.agents.create_thread()
    print(f"Created thread, thread ID {thread.id}")

    message = project_client.agents.create_message(thread_id=thread.id, role="user", content="Hello, tell me a joke")
    print(f"Created message, message ID {message.id}")

    with project_client.agents.create_stream(
        thread_id=thread.id, agent_id=agent.id, event_handler=MyEventHandler()
    ) as stream:
        for chunk in stream.get_stream_chunks():
            print(chunk)

    project_client.agents.delete_agent(agent.id)
    print("Deleted agent")

    messages = project_client.agents.list_messages(thread_id=thread.id)
    print(f"Messages: {messages}")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""
DESCRIPTION:
    Given an AIProjectClient, this sample demonstrates how to get an authenticated 
    AsyncAzureOpenAI client from the azure.ai.inference package.

USAGE:
    python sample_chat_completions_with_azure_openai_client_async.py

    Before running the sample:

    pip install azure-ai-projects aiohttp openai

    Set these environment variables with your own values:
    * PROJECT_CONNECTION_STRING - the Azure AI Project connection string, as found in your AI Foundry project.
    * MODEL_DEPLOYMENT_NAME - The model deployment name, as found in your AI Foundry project.

    Update the Azure OpenAI api-version as needed (see `api_version=` below). Values can be found here:
    https://learn.microsoft.com/azure/ai-services/openai/reference#api-specs
"""
import os
import asyncio
from azure.ai.projects.aio import AIProjectClient
from azure.identity.aio import DefaultAzureCredential


async def sample_get_azure_openai_client_async():

    project_connection_string = os.environ["PROJECT_CONNECTION_STRING"]
    model_deployment_name = os.environ["MODEL_DEPLOYMENT_NAME"]

    async with DefaultAzureCredential() as credential:

        async with AIProjectClient.from_connection_string(
            credential=credential,
            conn_str=project_connection_string,
        ) as project_client:

            # Get an authenticated AsyncAzureOpenAI client for your default Azure OpenAI connection:
            async with await project_client.inference.get_azure_openai_client(api_version="2024-06-01") as client:

                response = await client.chat.completions.create(
                    model=model_deployment_name,
                    messages=[
                        {
                            "role": "user",
                            "content": "How many feet are in a mile?",
                        },
                    ],
                )

                print(response.choices[0].message.content)


async def main():
    await sample_get_azure_openai_client_async()


if __name__ == "__main__":
    asyncio.run(main())
